import datetime as dt
from unittest.mock import Mock, patch
from urllib.parse import parse_qs, urlparse

from django.contrib.sessions.middleware import SessionMiddleware
from django.test import RequestFactory, TestCase
from django.urls import reverse
from django.utils.dateparse import parse_datetime
from django.utils.timezone import now

from allianceauth.eveonline.models import EveCharacter
from app_utils.testdata_factories import UserMainFactory
from app_utils.testing import json_response_to_python

import structures.views.status
from structures.models import Owner, Structure
from structures.tests.testdata.factories import (
    EveCharacterFactory,
    JumpGateFactory,
    OwnerFactory,
    PocoFactory,
    SkyhookFactory,
    StarbaseFactory,
    StructureFactory,
    StructureTagFactory,
    UserMainDefaultFactory,
    UserMainDefaultOwnerFactory,
    WebhookFactory,
)
from structures.tests.testdata.load_eveuniverse import load_eveuniverse
from structures.views import structures

from .utils import json_response_to_dict

VIEWS_PATH = "structures.views.structures"
OWNERS_PATH = "structures.models.owners"


@patch(VIEWS_PATH + ".STRUCTURES_DEFAULT_TAGS_FILTER_ENABLED", False)
class TestIndexRedirect(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()

    def test_should_redirect_to_public_view(self):
        # given
        user = UserMainFactory(
            permissions__=[
                "structures.basic_access",
            ],
        )
        request = self.factory.get("/")
        request.user = user

        # when
        response = structures.index(request)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("structures:public"))

    def test_should_redirect_to_structure_list_view_1(self):
        # given
        user = UserMainFactory(
            permissions__=[
                "structures.basic_access",
                "structures.view_corporation_structures",
            ],
        )
        request = self.factory.get("/")
        request.user = user
        # when
        response = structures.index(request)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("structures:structure_list"))

    def test_should_redirect_to_structure_list_view_2(self):
        # given
        user = UserMainFactory(
            permissions__=[
                "structures.basic_access",
                "structures.view_alliance_structures",
            ],
        )
        request = self.factory.get("/")
        request.user = user
        # when
        response = structures.index(request)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("structures:structure_list"))

    def test_should_redirect_to_structure_list_view_3(self):
        # given
        user = UserMainFactory(
            permissions__=[
                "structures.basic_access",
                "structures.view_all_structures",
            ],
        )
        request = self.factory.get("/")
        request.user = user
        # when
        response = structures.index(request)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("structures:structure_list"))


class TestIndexTagFilter(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        StructureTagFactory(name="tag_a", is_default=True)
        cls.user = UserMainDefaultFactory()

    @patch(VIEWS_PATH + ".STRUCTURES_DEFAULT_TAGS_FILTER_ENABLED", True)
    def test_default_filter_enabled(self):
        # given
        request = self.factory.get("/")
        request.user = self.user
        # when
        response = structures.index(request)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/structures/list?tags=tag_a")

    @patch(VIEWS_PATH + ".STRUCTURES_DEFAULT_TAGS_FILTER_ENABLED", False)
    def test_default_filter_disabled(self):
        # given
        request = self.factory.get("/")
        request.user = self.user
        # when
        response = structures.index(request)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/structures/list")


class TestStructureListDataFilterVariant(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        load_eveuniverse()
        cls.user = UserMainDefaultFactory()
        owner = OwnerFactory(user=cls.user)
        cls.structure = StructureFactory(owner=owner)
        cls.poco = PocoFactory(owner=owner)
        cls.skyhook = SkyhookFactory(owner=owner)
        cls.starbase = StarbaseFactory(owner=owner)
        cls.jump_gate = JumpGateFactory(owner=owner)

    def test_should_return_upwell_structures_only(self):
        # given
        request = self.factory.get("/")
        request.user = self.user
        # when
        response = structures.structure_list_data(request, "structures")
        # then
        self.assertEqual(response.status_code, 200)
        data = json_response_to_dict(response)
        structure_ids = set(data.keys())
        self.assertSetEqual(structure_ids, {self.structure.id, self.jump_gate.id})

    def test_should_return_orbitals_only(self):
        # given
        request = self.factory.get("/")
        request.user = self.user
        # when
        response = structures.structure_list_data(request, "orbitals")
        # then
        self.assertEqual(response.status_code, 200)
        data = json_response_to_dict(response)
        structure_ids = set(data.keys())
        self.assertSetEqual(structure_ids, {self.poco.id, self.skyhook.id})

    def test_should_return_starbases_only(self):
        # given
        request = self.factory.get("/")
        request.user = self.user
        # when
        response = structures.structure_list_data(request, "starbases")
        # then
        self.assertEqual(response.status_code, 200)
        data = json_response_to_dict(response)
        structure_ids = set(data.keys())
        self.assertSetEqual(structure_ids, {self.starbase.id})

    def test_should_return_jump_gates(self):
        # given
        request = self.factory.get("/")
        request.user = self.user
        # when
        response = structures.structure_list_data(request, "jump_gates")
        # then
        self.assertEqual(response.status_code, 200)
        data = json_response_to_dict(response)
        structure_ids = set(data.keys())
        self.assertSetEqual(structure_ids, {self.jump_gate.id})

    def test_should_return_all_structures(self):
        # given
        request = self.factory.get("/")
        request.user = self.user
        # when
        response = structures.structure_list_data(request, "all")
        # then
        self.assertEqual(response.status_code, 200)
        data = json_response_to_dict(response)
        structure_ids = set(data.keys())
        self.assertSetEqual(
            structure_ids,
            {
                self.structure.id,
                self.poco.id,
                self.starbase.id,
                self.jump_gate.id,
                self.skyhook.id,
            },
        )

    def test_should_raise_error_when_invalid_variant_requested(self):
        # given
        request = self.factory.get("/")
        request.user = self.user
        # when/then
        with self.assertRaises(ValueError):
            structures.structure_list_data(request, "invalid")

    def test_should_not_return_structure_from_different_corporations(self):
        # given
        other_structure = StructureFactory()
        request = self.factory.get("/")
        request.user = self.user
        # when
        response = structures.structure_list_data(request, "structures")
        # then
        self.assertEqual(response.status_code, 200)
        data = json_response_to_dict(response)
        structure_ids = set(data.keys())
        self.assertNotIn(other_structure.id, structure_ids)


class TestStructureListDataPermissions(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        load_eveuniverse()

    def test_should_show_structures_from_own_corporation_only(self):
        # given
        user = UserMainFactory(
            permissions__=[
                "structures.basic_access",
                "structures.view_corporation_structures",
            ],
        )
        owner = OwnerFactory(user=user)
        structure_own_corp = StructureFactory(owner=owner)
        StructureFactory()  # structure with a different corporation
        request = self.factory.get("/")
        request.user = user
        # when
        response = structures.structure_list_data(request, "all")
        # then
        self.assertEqual(response.status_code, 200)
        data = json_response_to_dict(response)
        structure_ids = set(data.keys())
        self.assertSetEqual(structure_ids, {structure_own_corp.id})

    def test_should_show_unanchoring_status(self):
        # given
        user = UserMainFactory(
            permissions__=[
                "structures.basic_access",
                "structures.view_all_structures",
                "structures.view_all_unanchoring_status",
            ],
        )
        owner = OwnerFactory(user=user)
        structure = StructureFactory(
            owner=owner, unanchors_at=now() + dt.timedelta(days=3)
        )
        request = self.factory.get("/")
        request.user = user
        # when
        response = structures.structure_list_data(request, "all")
        # then
        self.assertEqual(response.status_code, 200)
        data = json_response_to_dict(response)
        structure = data[structure.id]
        self.assertIn("Unanchoring until", structure["state_details"])

    def test_should_not_show_unanchoring_status(self):
        # given
        user = UserMainFactory(
            permissions__=["structures.basic_access", "structures.view_all_structures"],
        )
        owner = OwnerFactory(user=user)
        structure = StructureFactory(
            owner=owner, unanchors_at=now() + dt.timedelta(days=3)
        )
        request = self.factory.get("/")
        request.user = user
        # when
        response = structures.structure_list_data(request, "all")
        # then
        self.assertEqual(response.status_code, 200)
        data = json_response_to_dict(response)
        structure = data[structure.id]
        self.assertNotIn("Unanchoring until", structure["state_details"])


class TestStructureListTagFilters(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        cls.factory = RequestFactory()
        cls.user = UserMainDefaultFactory()
        cls.owner = OwnerFactory(user=cls.user)
        tag_b = StructureTagFactory(name="tag_b")
        tag_c = StructureTagFactory(name="tag_c")
        cls.structure_1 = StructureFactory(
            owner=cls.owner, tags=[tag_b], id=1000000000001
        )
        cls.structure_2 = StructureFactory(
            owner=cls.owner, tags=[tag_c], id=1000000000002
        )
        cls.structure_3 = StructureFactory(
            owner=cls.owner, tags=[tag_b, tag_c], id=1000000000003
        )
        cls.structure_4 = StructureFactory(owner=cls.owner, id=1000000000004)

    def test_list_filter_by_tag_with_no_filter(self):
        # given
        request = self.factory.get("/")
        request.user = self.user
        # when
        response = structures.structure_list_data(request, "all")
        # then
        self.assertEqual(response.status_code, 200)
        data = json_response_to_dict(response)
        self.assertSetEqual(
            set(data.keys()),
            {
                self.structure_1.id,
                self.structure_2.id,
                self.structure_3.id,
                self.structure_4.id,
            },
        )

    def test_list_filter_by_one_tag(self):
        # given
        request = self.factory.get("/?tags=tag_b")
        request.user = self.user
        # when
        response = structures.structure_list_data(request, "all")
        # then
        self.assertEqual(response.status_code, 200)
        data = json_response_to_dict(response)
        self.assertSetEqual(
            set(data.keys()), {self.structure_1.id, self.structure_3.id}
        )

    def test_list_filter_by_two_tags(self):
        # given
        request = self.factory.get("/?tags=tag_b,tag_c")
        request.user = self.user
        # when
        response = structures.structure_list_data(request, "all")
        # then
        self.assertEqual(response.status_code, 200)
        data = json_response_to_dict(response)
        self.assertSetEqual(
            set(data.keys()),
            {self.structure_1.id, self.structure_2.id, self.structure_3.id},
        )

    def test_call_with_raw_tags(self):
        # given
        request = self.factory.get("/?tags=tag_c,tag_b")
        request.user = self.user
        # when
        response = structures.structure_list(request)
        # then
        self.assertEqual(response.status_code, 200)

    def test_set_tags_filter(self):
        # given
        request = self.factory.post("/", data={"tag_b": True, "tag_c": True})
        request.user = self.user
        # when
        response = structures.structure_list(request)
        # then
        self.assertEqual(response.status_code, 302)
        parts = urlparse(response.url)
        path = parts[2]
        query_dict = parse_qs(parts[4])
        self.assertEqual(path, reverse("structures:structure_list"))
        self.assertIn("tags", query_dict)
        params = query_dict["tags"][0].split(",")
        self.assertSetEqual(set(params), {"tag_c", "tag_b"})

    def test_handle_post_with_no_tags(self):
        # given
        request = self.factory.post("/")
        request.user = self.user
        # when
        response = structures.structure_list(request)
        # then
        self.assertEqual(response.status_code, 302)
        parts = urlparse(response.url)
        path = parts[2]
        query_dict = parse_qs(parts[4])
        self.assertEqual(path, reverse("structures:structure_list"))
        self.assertNotIn("tags", query_dict)


class TestStructurePowerModes(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        load_eveuniverse()
        cls.user = UserMainDefaultOwnerFactory()
        cls.owner = OwnerFactory(user=cls.user)

    def display_data_for_structure(self, structure_id: int):
        request = self.factory.get("/")
        request.user = self.user
        response = structures.structure_list_data(request, "all")
        self.assertEqual(response.status_code, 200)

        data = json_response_to_python(response)["data"]
        for row in data:
            if row["id"] == structure_id:
                return row

        return None

    def test_full_power(self):
        # given
        structure = StructureFactory(
            owner=self.owner, fuel_expires_at=now() + dt.timedelta(hours=1)
        )
        # when
        obj = self.display_data_for_structure(structure.id)
        # then
        self.assertEqual(obj["power_mode_str"], "Full Power")
        self.assertEqual(
            parse_datetime(obj["fuel_and_power"]["fuel_expires_at"]),
            structure.fuel_expires_at,
        )
        self.assertIn("Full Power", obj["fuel_and_power"]["display"])

    def test_low_power(self):
        # given
        structure = StructureFactory(
            owner=self.owner,
            fuel_expires_at=None,
            last_online_at=now() - dt.timedelta(days=3),
        )
        # when
        obj = self.display_data_for_structure(structure.id)
        # then
        self.assertEqual(obj["power_mode_str"], "Low Power")
        self.assertIn("Low Power", obj["fuel_and_power"]["display"])

    def test_abandoned(self):
        # given
        structure = StructureFactory(
            owner=self.owner,
            fuel_expires_at=None,
            last_online_at=now() - dt.timedelta(days=7, seconds=1),
        )
        # when
        obj = self.display_data_for_structure(structure.id)
        # then
        self.assertEqual(obj["power_mode_str"], "Abandoned")
        self.assertIn("Abandoned", obj["fuel_and_power"]["display"])

    def test_maybe_abandoned(self):
        # given
        structure = StructureFactory(
            owner=self.owner, fuel_expires_at=None, last_online_at=None
        )
        # when
        obj = self.display_data_for_structure(structure.id)
        # then
        self.assertEqual(obj["power_mode_str"], "Abandoned?")
        self.assertIn("Abandoned?", obj["fuel_and_power"]["display"])

    def test_poco(self):
        # given
        structure = PocoFactory(owner=self.owner)
        # when
        obj = self.display_data_for_structure(structure.id)
        self.assertEqual(obj["power_mode_str"], "")
        self.assertEqual(obj["fuel_and_power"]["display"], "")

    def test_starbase_online(self):
        # given
        structure = StarbaseFactory(
            owner=self.owner, fuel_expires_at=now() + dt.timedelta(hours=1)
        )
        # when
        obj = self.display_data_for_structure(structure.id)
        self.assertEqual(obj["power_mode_str"], "")
        self.assertEqual(
            parse_datetime(obj["fuel_and_power"]["fuel_expires_at"]),
            structure.fuel_expires_at,
        )

    def test_starbase_offline(self):
        # given
        structure = StarbaseFactory(
            owner=self.owner, fuel_expires_at=None, state=Structure.State.POS_OFFLINE
        )
        # when
        obj = self.display_data_for_structure(structure.id)
        # then
        self.assertEqual(obj["power_mode_str"], "")
        self.assertIn("-", obj["fuel_and_power"]["display"])


class TestAddStructureOwner(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        load_eveuniverse()
        cls.user = UserMainDefaultOwnerFactory()
        cls.character: EveCharacter = cls.user.profile.main_character
        cls.character_ownership = cls.character.character_ownership

    def _add_structure_owner_view(self, token=None, user=None):
        # given
        request = self.factory.get(reverse("structures:add_structure_owner"))
        if not user:
            user = self.user
        if not token:
            token = user.token_set.first()
        request.user = user
        request.token = token
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        orig_view = structures.add_structure_owner.__wrapped__.__wrapped__.__wrapped__
        # when
        return orig_view(request, token)

    @patch(VIEWS_PATH + ".STRUCTURES_ADMIN_NOTIFICATIONS_ENABLED", True)
    @patch(VIEWS_PATH + ".tasks.update_all_for_owner")
    @patch(VIEWS_PATH + ".notify_admins")
    @patch(VIEWS_PATH + ".messages")
    def test_should_add_new_owner_and_notify_admins(
        self, mock_messages, mock_notify_admins, mock_update_all_for_owner
    ):
        # given
        webhook = WebhookFactory(is_default=True)
        # when
        response = self._add_structure_owner_view()
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("structures:index"))
        self.assertTrue(mock_messages.info.called)
        self.assertTrue(mock_notify_admins.called)
        new_owner = Owner.objects.first()
        self.assertSetEqual(
            {self.character_ownership.pk},
            set(new_owner.characters.values_list("character_ownership", flat=True)),
        )
        self.assertIn(webhook, new_owner.webhooks.all())
        self.assertTrue(mock_update_all_for_owner.delay.called)

    @patch(VIEWS_PATH + ".STRUCTURES_ADMIN_NOTIFICATIONS_ENABLED", False)
    @patch(VIEWS_PATH + ".tasks.update_all_for_owner")
    @patch(VIEWS_PATH + ".notify_admins")
    @patch(VIEWS_PATH + ".messages")
    def test_should_add_new_owner_and_not_notify_admins(
        self, mock_messages, mock_notify_admins, mock_update_all_for_owner
    ):
        # when
        response = self._add_structure_owner_view()
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("structures:index"))
        owner = Owner.objects.first()
        self.assertSetEqual(
            {self.character_ownership.pk},
            set(owner.characters.values_list("character_ownership", flat=True)),
        )
        self.assertTrue(mock_messages.info.called)
        self.assertFalse(mock_notify_admins.called)
        self.assertTrue(mock_update_all_for_owner.delay.called)

    @patch(VIEWS_PATH + ".STRUCTURES_ADMIN_NOTIFICATIONS_ENABLED", False)
    @patch(VIEWS_PATH + ".tasks.update_all_for_owner")
    @patch(VIEWS_PATH + ".notify_admins")
    @patch(VIEWS_PATH + ".messages")
    def test_should_add_structure_owner_with_no_default_webhook(
        self, mock_messages, mock_notify_admins, mock_update_all_for_owner
    ):
        # given
        WebhookFactory(is_default=False)
        # when
        response = self._add_structure_owner_view()
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("structures:index"))
        self.assertTrue(mock_messages.info.called)
        self.assertFalse(mock_notify_admins.called)
        new_owner = Owner.objects.get(
            characters__character_ownership=self.character_ownership
        )
        self.assertFalse(new_owner.webhooks.exists())
        self.assertTrue(mock_update_all_for_owner.delay.called)

    @patch(VIEWS_PATH + ".messages")
    def test_should_report_error_when_token_does_not_belong_to_user(
        self, mock_messages
    ):
        # given
        other_user = UserMainDefaultOwnerFactory()
        # when
        my_token = other_user.token_set.first()
        response = self._add_structure_owner_view(token=my_token)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("structures:index"))
        self.assertTrue(mock_messages.error.called)

    @patch(VIEWS_PATH + ".STRUCTURES_ADMIN_NOTIFICATIONS_ENABLED", False)
    @patch(VIEWS_PATH + ".tasks.update_all_for_owner")
    @patch(VIEWS_PATH + ".notify_admins")
    @patch(VIEWS_PATH + ".messages")
    def test_should_add_another_character_to_existing_owner_and_reactivate(
        self, mock_messages, mock_notify_admins, mock_update_all_for_owner
    ):
        # given
        owner = OwnerFactory(
            user=self.user, characters=[self.character], is_active=False
        )
        character_2 = EveCharacterFactory(corporation=owner.corporation)
        user_2 = UserMainDefaultOwnerFactory(main_character__character=character_2)
        # when
        response = self._add_structure_owner_view(user=user_2)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("structures:index"))
        self.assertTrue(mock_messages.info.called)
        self.assertFalse(mock_update_all_for_owner.delay.called)
        owner.refresh_from_db()
        character_ownerships = set(
            owner.characters.values_list("character_ownership", flat=True)
        )
        self.assertSetEqual(
            {self.character_ownership.pk, character_2.character_ownership.pk},
            character_ownerships,
        )
        self.assertTrue(owner.is_active)

    @patch(VIEWS_PATH + ".STRUCTURES_ADMIN_NOTIFICATIONS_ENABLED", False)
    @patch(VIEWS_PATH + ".tasks.update_all_for_owner")
    @patch(VIEWS_PATH + ".notify_admins")
    @patch(VIEWS_PATH + ".messages")
    def test_can_readd_same_character(
        self, mock_messages, mock_notify_admins, mock_update_all_for_owner
    ):
        # given
        owner = OwnerFactory(characters=[self.character])
        owner_character = owner.characters.first()
        # when
        response = self._add_structure_owner_view(user=self.user)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("structures:index"))
        owner.refresh_from_db()
        character_ownerships = set(
            owner.characters.values_list("character_ownership", flat=True)
        )
        self.assertSetEqual(
            {self.character_ownership.pk, owner_character.character_ownership.pk},
            character_ownerships,
        )

    # @patch(VIEWS_PATH + ".STRUCTURES_ADMIN_NOTIFICATIONS_ENABLED", False)
    # @patch(VIEWS_PATH + ".tasks.update_all_for_owner")
    # @patch(VIEWS_PATH + ".notify_admins")
    # @patch(VIEWS_PATH + ".messages")
    # def test_should_reenable_character_when_re_adding(
    #     self, mock_messages, mock_notify_admins, mock_update_all_for_owner
    # ):
    #     # given
    #     owner = OwnerFactory(characters=[self.character])
    #     owner_character: OwnerCharacter = owner.characters.first()
    #     owner_character.is_enabled = False
    #     owner_character.save()
    #     # when
    #     response = self._add_structure_owner_view(user=self.user)
    #     # then
    #     self.assertEqual(response.status_code, 302)
    #     self.assertEqual(response.url, reverse("structures:index"))
    #     owner.refresh_from_db()
    #     character_ownerships = set(
    #         owner.characters.values_list("character_ownership", flat=True)
    #     )
    #     self.assertSetEqual(
    #         {self.character_ownership.pk, owner_character.character_ownership.pk},
    #         character_ownerships,
    #     )
    #     owner_character.refresh_from_db()
    #     self.assertTrue(owner_character.is_enabled)


class TestStructureFittingModal(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        load_eveuniverse()
        cls.character = EveCharacterFactory()
        owner = OwnerFactory(corporation=cls.character.corporation)
        cls.structure = StructureFactory(owner=owner)

    def test_should_have_access_to_fitting(self):
        # given
        user = UserMainFactory(
            main_character__character=self.character,
            permissions__=[
                "structures.basic_access",
                "structures.view_corporation_structures",
                "structures.view_structure_fit",
            ],
        )
        request = self.factory.get("/")
        request.user = user
        # when
        response = structures.structure_details(request, self.structure.id)
        # then
        self.assertEqual(response.status_code, 200)

    def test_should_not_have_access_to_fitting(self):
        # given
        user = UserMainFactory(
            main_character__character=self.character,
            permissions__=[
                "structures.basic_access",
                "structures.view_corporation_structures",
            ],
        )
        request = self.factory.get("/")
        request.user = user
        # when
        response = structures.structure_details(request, self.structure.id)
        # then
        self.assertEqual(response.status_code, 302)


class TestDetailsModal(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        load_eveuniverse()
        cls.user = UserMainDefaultFactory()
        cls.owner = OwnerFactory(user=cls.user)

    def test_should_load_poco_detail(self):
        # given
        structure = PocoFactory(owner=self.owner)
        request = self.factory.get("/")
        request.user = self.user
        # when
        response = structures.poco_details(request, structure.id)
        # then
        self.assertEqual(response.status_code, 200)

    def test_should_load_starbase_detail(self):
        # given
        structure = StarbaseFactory(owner=self.owner)
        request = self.factory.get("/")
        request.user = self.user
        # when
        response = structures.starbase_detail(request, structure.id)
        # then
        self.assertEqual(response.status_code, 200)



"""Helpers for Structures."""

import datetime as dt
from typing import Any, List, Optional, Union
from urllib.parse import urlparse

from django.utils.html import format_html, format_html_join
from django.utils.safestring import SafeText, mark_safe
from django.utils.timezone import now
from eveuniverse.models import EveEntity, EveType


def hours_until_deadline(
    deadline: dt.datetime, start: Optional[dt.datetime] = None
) -> float:
    """Currently remaining hours until a given deadline."""
    if not isinstance(deadline, dt.datetime):
        raise TypeError("deadline must be of type datetime")
    if not start:
        start = now()
    return (deadline - start).total_seconds() / 3600


def datetime_almost_equal(
    first: Optional[dt.datetime], second: Optional[dt.datetime], threshold: int
) -> bool:
    """True when first and second datetime are within threshold in seconds.
    False when first or second is None.
    """
    if not first or not second:
        return False
    dif = abs((first - second).total_seconds())
    return dif <= abs(threshold)


def is_absolute_url(url: str) -> bool:
    """Return True if URL is absolute else False."""
    return bool(urlparse(url).netloc)


def get_or_create_esi_obj(model_class: type, *args, **kwargs) -> Any:
    """Get or create an object from ESI and return it."""
    obj, _ = model_class.objects.get_or_create_esi(*args, **kwargs)
    return obj


def get_or_create_eve_entity(*args, **kwargs) -> EveEntity:
    """Get or create an EveEntity object from ESI and return it."""
    obj = get_or_create_esi_obj(EveEntity, *args, **kwargs)
    return obj


def get_or_create_eve_type(*args, **kwargs) -> EveType:
    """Get or create an EveEntity object from ESI and return it."""
    obj = get_or_create_esi_obj(EveType, *args, **kwargs)
    return obj


def floating_icon_with_text_html(
    icon_url: str, lines: List[Union[str, SafeText]]
) -> SafeText:
    """Return HTML for a multi-line paragraph with a floating icon on the left.

    HTML in lines will not be escaped if they are marked as safe.
    """
    icon_html = format_html(('<img src="{}" class="floating-icon">'), icon_url)
    text_html = format_html_join(mark_safe("<br>"), "{}", ((line,) for line in lines))
    result_html = format_html("<p>{}{}</p>", icon_html, text_html)
    return result_html



import datetime as dt
from unittest.mock import Mock, patch

import pytz

from django.utils.timezone import now
from eveuniverse.models import EveType

from app_utils.django import app_labels
from app_utils.testing import NoSocketsTestCase

from structures.constants import EveTypeId
from structures.core import notification_timers
from structures.core.notification_types import NotificationType
from structures.models import Structure
from structures.tests.testdata.factories import (
    EveEntityAllianceFactory,
    EveEntityCharacterFactory,
    GeneratedNotificationFactory,
    NotificationFactory,
    NotificationMoonMiningExtractionCanceledFactory,
    NotificationMoonMiningExtractionStartedFactory,
    NotificationOrbitalReinforcedFactory,
    NotificationSovStructureReinforcedFactory,
    NotificationStructureLostShieldFactory,
    OwnerFactory,
    PocoFactory,
    RefineryFactory,
    StructureFactory,
)
from structures.tests.testdata.load_eveuniverse import load_eveuniverse

if "structuretimers" in app_labels():
    from structuretimers.models import Timer

    MODULE_PATH = "structures.core.notification_timers"

    @patch(
        "structuretimers.models._task_calc_timer_distances_for_all_staging_systems",
        Mock(),
    )
    @patch("structuretimers.models.STRUCTURETIMERS_NOTIFICATIONS_ENABLED", False)
    @patch(MODULE_PATH + ".STRUCTURES_MOON_EXTRACTION_TIMERS_ENABLED", True)
    class TestTimersForStructureTimers(NoSocketsTestCase):
        @classmethod
        def setUpClass(cls):
            super().setUpClass()
            load_eveuniverse()

        def test_should_create_timer_for_reinforced_structure(self):
            # given
            owner = OwnerFactory()
            structure = StructureFactory(owner=owner)
            notif = NotificationStructureLostShieldFactory(
                owner=owner, structure=structure
            )
            # when
            result = notification_timers.add_or_remove_timer(notif)
            # then
            self.assertTrue(result)
            timer = Timer.objects.first()
            self.assertIsInstance(timer, Timer)
            self.assertEqual(timer.eve_solar_system, structure.eve_solar_system)
            self.assertEqual(timer.structure_type, structure.eve_type)
            self.assertEqual(timer.timer_type, Timer.Type.ARMOR)
            self.assertEqual(timer.objective, Timer.Objective.FRIENDLY)
            self.assertAlmostEqual(
                timer.date, now() + dt.timedelta(hours=47), delta=dt.timedelta(hours=1)
            )
            self.assertEqual(timer.eve_corporation, owner.corporation)
            self.assertEqual(timer.eve_alliance, owner.corporation.alliance)

            self.assertEqual(timer.visibility, Timer.Visibility.UNRESTRICTED)
            self.assertEqual(timer.structure_name, structure.name)
            self.assertEqual(timer.owner_name, owner.corporation.corporation_name)
            self.assertTrue(timer.details_notes)
            notif.refresh_from_db()
            self.assertTrue(notif.is_timer_added)

        def test_should_create_timer_for_sov_reinforcement(self):
            # given
            owner = OwnerFactory(is_alliance_main=True)
            alliance = owner.corporation.alliance
            sender = EveEntityAllianceFactory(
                id=alliance.alliance_id, name=alliance.alliance_name
            )
            notif = NotificationSovStructureReinforcedFactory(
                owner=owner, sender=sender, eve_solar_system_name="1-PGSG"
            )
            # when
            result = notification_timers.add_or_remove_timer(notif)
            # then
            self.assertTrue(result)
            timer = Timer.objects.first()
            self.assertIsInstance(timer, Timer)
            self.assertEqual(timer.timer_type, Timer.Type.FINAL)
            self.assertEqual(timer.eve_solar_system.id, 30000474)
            self.assertEqual(timer.structure_type.id, EveTypeId.TCU)
            self.assertAlmostEqual(
                timer.date,
                pytz.utc.localize(dt.datetime(2018, 12, 20, 17, 3, 22)),
                delta=dt.timedelta(seconds=120),
            )
            self.assertEqual(timer.eve_corporation, owner.corporation)
            self.assertEqual(timer.eve_alliance, owner.corporation.alliance)
            self.assertEqual(timer.visibility, Timer.Visibility.UNRESTRICTED)
            self.assertEqual(timer.owner_name, alliance.alliance_name)
            self.assertTrue(timer.details_notes)
            notif.refresh_from_db()
            self.assertTrue(notif.is_timer_added)

        def test_should_create_timer_for_sov_reinforcement_2(self):
            # given
            owner = OwnerFactory(is_alliance_main=False)
            alliance = owner.corporation.alliance
            sender = EveEntityAllianceFactory(
                id=alliance.alliance_id, name=alliance.alliance_name
            )
            notif = NotificationSovStructureReinforcedFactory(
                owner=owner, sender=sender, eve_solar_system_name="1-PGSG"
            )
            # when
            result = notification_timers.add_or_remove_timer(notif)
            # then
            self.assertFalse(result)
            self.assertFalse(Timer.objects.exists())
            notif.refresh_from_db()
            self.assertFalse(notif.is_timer_added)

        def test_should_create_timer_for_orbital_reinforcements(self):
            # given
            owner = OwnerFactory()
            structure = PocoFactory(owner=owner)
            notif = NotificationOrbitalReinforcedFactory(
                owner=owner, structure=structure
            )
            # when
            result = notification_timers.add_or_remove_timer(notif)
            # then
            self.assertTrue(result)
            timer = Timer.objects.first()
            self.assertIsInstance(timer, Timer)
            self.assertEqual(timer.timer_type, Timer.Type.FINAL)
            self.assertEqual(timer.eve_solar_system.id, structure.eve_solar_system.id)
            self.assertEqual(
                timer.structure_type, EveType.objects.get(id=EveTypeId.CUSTOMS_OFFICE)
            )
            self.assertEqual(timer.location_details, structure.eve_planet.name)
            self.assertAlmostEqual(
                timer.date,
                pytz.utc.localize(dt.datetime(2019, 10, 13, 20, 32, 27)),
                delta=dt.timedelta(seconds=120),
            )
            self.assertEqual(timer.eve_corporation, owner.corporation)
            self.assertEqual(timer.eve_alliance, owner.corporation.alliance)
            self.assertEqual(timer.visibility, Timer.Visibility.UNRESTRICTED)
            self.assertEqual(timer.owner_name, owner.corporation.corporation_name)
            self.assertTrue(timer.details_notes)
            notif.refresh_from_db()
            self.assertTrue(notif.is_timer_added)

        def test_should_create_timer_for_moon_extraction(self):
            # given
            owner = OwnerFactory()
            structure = RefineryFactory(owner=owner)
            started_by = EveEntityCharacterFactory()
            notif = NotificationMoonMiningExtractionStartedFactory(
                owner=owner, structure=structure, started_by=started_by
            )
            # when
            result = notification_timers.add_or_remove_timer(notif)
            # then
            self.assertTrue(result)
            timer = Timer.objects.first()
            self.assertIsInstance(timer, Timer)
            self.assertEqual(timer.timer_type, Timer.Type.MOONMINING)
            self.assertEqual(timer.eve_solar_system, structure.eve_solar_system)
            self.assertEqual(timer.structure_type, structure.eve_type)
            self.assertEqual(timer.eve_corporation, owner.corporation)
            self.assertEqual(timer.eve_alliance, owner.corporation.alliance)
            self.assertEqual(timer.visibility, Timer.Visibility.UNRESTRICTED)
            self.assertEqual(timer.location_details, structure.eve_moon.name)
            self.assertEqual(timer.owner_name, owner.corporation.corporation_name)
            self.assertEqual(timer.structure_name, structure.name)
            self.assertTrue(timer.details_notes)
            notif.refresh_from_db()
            self.assertTrue(notif.is_timer_added)

        def test_can_delete_extraction_timer(self):
            # create timer
            owner = OwnerFactory()
            structure = RefineryFactory(owner=owner)
            notif = NotificationMoonMiningExtractionStartedFactory(
                owner=owner, structure=structure
            )
            self.assertTrue(notification_timers.add_or_remove_timer(notif))
            timer = Timer.objects.first()
            self.assertIsInstance(timer, Timer)
            notif.refresh_from_db()
            self.assertTrue(notif.is_timer_added)

            # delete timer
            notif = NotificationMoonMiningExtractionCanceledFactory(
                owner=owner, structure=structure
            )
            self.assertTrue(notification_timers.add_or_remove_timer(notif))
            self.assertFalse(Timer.objects.filter(pk=timer.pk).exists())
            notif.refresh_from_db()
            self.assertTrue(notif.is_timer_added)

        def test_should_create_timer_for_starbase_reinforcement(self):
            # given
            notif = GeneratedNotificationFactory()
            structure: Structure = notif.structures.first()
            # when
            result = notification_timers.add_or_remove_timer(notif)
            # then
            self.assertTrue(result)
            obj = Timer.objects.first()
            self.assertEqual(obj.eve_solar_system, structure.eve_solar_system)
            self.assertEqual(obj.structure_type, structure.eve_type)
            self.assertEqual(obj.timer_type, Timer.Type.FINAL)
            self.assertEqual(obj.objective, Timer.Objective.FRIENDLY)
            self.assertAlmostEqual(obj.date, structure.state_timer_end)
            self.assertEqual(obj.eve_corporation, structure.owner.corporation)
            self.assertEqual(obj.visibility, Timer.Visibility.UNRESTRICTED)
            self.assertEqual(obj.structure_name, structure.name)
            self.assertEqual(
                obj.owner_name, structure.owner.corporation.corporation_name
            )
            self.assertTrue(obj.details_notes)
            notif.refresh_from_db()
            self.assertTrue(notif.is_timer_added)

        def test_raise_error_for_unsupported_types(self):
            owner = OwnerFactory()
            notif = NotificationFactory(
                owner=owner, notif_type=NotificationType.WAR_CORPORATION_BECAME_ELIGIBLE
            )
            # when
            with self.assertRaises(NotImplementedError):
                notification_timers.add_or_remove_timer(notif)



# Generated by Django 4.0.10 on 2024-08-19 18:59

import multiselectfield.db.fields

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("structures", "0007_add_notificationtypes_skyhook_metenox"),
    ]

    operations = [
        migrations.AlterField(
            model_name="webhook",
            name="notification_types",
            field=multiselectfield.db.fields.MultiSelectField(
                choices=[
                    ("OwnershipTransferred", "Upwell structure ownership transferred"),
                    ("StructureAnchoring", "Upwell structure anchoring"),
                    ("StructureDestroyed", "Upwell structure destroyed"),
                    ("StructureFuelAlert", "Upwell structure fuel alert"),
                    ("StructureJumpFuelAlert", "Upwell structure jump fuel alert"),
                    ("StructureLostArmor", "Upwell structure lost armor"),
                    ("StructureLostShields", "Upwell structure lost shields"),
                    ("StructureLowReagentsAlert", "Structure low reagents alert"),
                    ("StructureNoReagentsAlert", "Structure no reagents alert"),
                    ("StructureOnline", "Upwell structure went online"),
                    ("StructureRefueledExtra", "Upwell structure refueled"),
                    (
                        "StructuresReinforcementChanged",
                        "Upwell structure reinforcement time changed",
                    ),
                    (
                        "StructureServicesOffline",
                        "Upwell structure services went offline",
                    ),
                    ("StructureUnanchoring", "Upwell structure unanchoring"),
                    ("StructureUnderAttack", "Upwell structure is under attack"),
                    ("StructureWentHighPower", "Upwell structure went high power"),
                    ("StructureWentLowPower", "Upwell structure went low power"),
                    ("SkyhookDeployed", "Skyhook deployed"),
                    ("SkyhookDestroyed", "Skyhook destroyed"),
                    ("SkyhookLostShields", "Skyhook lost shields"),
                    ("SkyhookOnline", "Skyhook online"),
                    ("SkyhookUnderAttack", "Skyhook under attack"),
                    ("OrbitalAttacked", "Customs office attacked"),
                    ("OrbitalReinforced", "Customs office reinforced"),
                    ("TowerAlertMsg", "Starbase attacked"),
                    ("TowerRefueledExtra", "Starbase refueled (BETA)"),
                    ("TowerReinforcedExtra", "Starbase reinforced (BETA)"),
                    ("TowerResourceAlertMsg", "Starbase fuel alert"),
                    (
                        "MoonminingAutomaticFracture",
                        "Moon mining automatic fracture triggered",
                    ),
                    (
                        "MoonminingExtractionCancelled",
                        "Moon mining extraction cancelled",
                    ),
                    ("MoonminingExtractionFinished", "Moon mining extraction finished"),
                    ("MoonminingExtractionStarted", "Moon mining extraction started"),
                    ("MoonminingLaserFired", "Moonmining laser fired"),
                    ("SovAllClaimAquiredMsg", "Sovereignty claim acknowledgment"),
                    ("SovAllClaimLostMsg", "Sovereignty lost"),
                    ("AllAnchoringMsg", "Structure anchoring in alliance space"),
                    ("EntosisCaptureStarted", "Sovereignty entosis capture started"),
                    ("SovStructureDestroyed", "Sovereignty structure destroyed"),
                    ("SovStructureReinforced", "Sovereignty structure reinforced"),
                    (
                        "SovCommandNodeEventStarted",
                        "Sovereignty command node event started",
                    ),
                    ("AcceptedAlly", "War accepted ally"),
                    (
                        "AllWarCorpJoinedAllianceMsg",
                        "Alliance war corporation joined alliance message",
                    ),
                    ("AllWarSurrenderMsg", "Alliance war surrender message"),
                    ("AllyJoinedWarAggressorMsg", "War ally joined aggressor"),
                    ("AllyJoinedWarAllyMsg", "War ally joined ally"),
                    ("AllyJoinedWarDefenderMsg", "War ally joined defender"),
                    ("CorpWarSurrenderMsg", "War party surrendered"),
                    ("CorpBecameWarEligible", "War corporation became eligible"),
                    ("CorpNoLongerWarEligible", "War corporation no longer eligible"),
                    ("DeclareWar", "War declared"),
                    ("WarHQRemovedFromSpace", "War HQ removed from space"),
                    ("WarInvalid", "War invalid"),
                    (
                        "MercOfferedNegotiationMsg",
                        "War mercenary offered negotiation message",
                    ),
                    ("MercOfferRetractedMsg", "War mercenary offer retracted message"),
                    ("OfferedSurrender", "War offered surrender"),
                    ("OfferedToAlly", "War offered to become ally"),
                    ("WarAdopted", "War adopted"),
                    ("WarDeclared", "War declared"),
                    ("WarInherited", "War inherited"),
                    ("WarRetractedByConcord", "War retracted by Concord"),
                    ("WarSurrenderOfferMsg", "War surrender offered"),
                    ("CharAppAcceptMsg", "Character joins corporation"),
                    ("CorpAppInvitedMsg", "Character invited to join corporation"),
                    ("CorpAppNewMsg", "Character submitted application"),
                    ("CharAppRejectMsg", "Corp application rejected message"),
                    (
                        "CorpAppRejectCustomMsg",
                        "Corp application rejected custom message",
                    ),
                    ("CharAppWithdrawMsg", "Character withdrew application"),
                    ("CharLeftCorpMsg", "Character leaves corporation"),
                    ("BillOutOfMoneyMsg", "Bill out of money"),
                    ("CorpAllBillMsg", "Corp alliance billing message"),
                    (
                        "InfrastructureHubBillAboutToExpire",
                        "I-HUB bill about to expire",
                    ),
                    ("IHubDestroyedByBillFailure", "I_HUB destroyed by bill failure"),
                ],
                default=[
                    "OrbitalAttacked",
                    "OrbitalReinforced",
                    "SkyhookDestroyed",
                    "SkyhookLostShields",
                    "SkyhookOnline",
                    "SkyhookUnderAttack",
                    "SovStructureDestroyed",
                    "SovStructureReinforced",
                    "StructureAnchoring",
                    "StructureDestroyed",
                    "StructureFuelAlert",
                    "StructureLostArmor",
                    "StructureLostShields",
                    "StructureLowReagentsAlert",
                    "StructureNoReagentsAlert",
                    "StructureOnline",
                    "StructureServicesOffline",
                    "StructureUnderAttack",
                    "StructureWentHighPower",
                    "StructureWentLowPower",
                    "TowerAlertMsg",
                    "TowerResourceAlertMsg",
                ],
                help_text="Select which type of notifications should be forwarded to this webhook",
                max_length=1474,
                verbose_name="notification types",
            ),
        ),
    ]



import datetime as dt

import dhooks_lite

from django.test import TestCase, override_settings
from django.utils.timezone import now

from app_utils.testing import NoSocketsTestCase

from structures.core.notification_embeds import NotificationBaseEmbed
from structures.core.notification_embeds.billing_embeds import BillType
from structures.core.notification_embeds.tower_embeds import (
    NotificationTowerReinforcedExtra,
)
from structures.core.notification_embeds.war_embeds import (
    NotificationWarCorporationBecameEligible,
)
from structures.core.notification_types import NotificationType
from structures.models.notifications import Notification, Webhook
from structures.tests.testdata.factories import (
    EveAllianceInfoFactory,
    EveCorporationInfoFactory,
    EveEntityAllianceFactory,
    EveEntityCorporationFactory,
    GeneratedNotificationFactory,
    NotificationFactory,
    OwnerFactory,
    StarbaseFactory,
    StructureFactory,
    UserMainDefaultOwnerFactory,
)
from structures.tests.testdata.helpers import (
    load_eve_entities,
    load_notification_entities,
    markdown_to_plain,
)
from structures.tests.testdata.load_eveuniverse import load_eveuniverse

MODULE_PATH = "structures.core.notification_embeds"


class TestBilType(TestCase):
    def test_should_create_from_valid_id(self):
        self.assertEqual(BillType.to_enum(7), BillType.INFRASTRUCTURE_HUB)

    def test_should_create_from_invalid_id(self):
        for bill_id in range(7):
            with self.subTest(bill_id=bill_id):
                self.assertEqual(BillType.to_enum(bill_id), BillType.UNKNOWN)


class TestNotificationEmbeds(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.owner = OwnerFactory()

    def test_should_create_obj_from_notification(self):
        # given
        notification = NotificationFactory(
            owner=self.owner,
            notification_id=1000000999,
            notif_type=NotificationType.WAR_CORPORATION_BECAME_ELIGIBLE,
        )
        # when
        notification_embed = NotificationBaseEmbed.create(notification)
        # then
        self.assertIsInstance(
            notification_embed, NotificationWarCorporationBecameEligible
        )
        self.assertEqual(str(notification_embed), "1000000999:CorpBecameWarEligible")

    def test_should_require_notification_for_init(self):
        with self.assertRaises(TypeError):
            NotificationBaseEmbed(notification="dummy")

    def test_should_require_notification_for_factory(self):
        with self.assertRaises(TypeError):
            NotificationBaseEmbed.create(notification="dummy")


class TestNotificationEmbedsGenerate(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_eve_entities()
        alliance = EveAllianceInfoFactory(alliance_id=3001)
        corporation = EveCorporationInfoFactory(corporation_id=2001, alliance=alliance)
        cls.owner = OwnerFactory(corporation=corporation)
        load_notification_entities(cls.owner)

    def test_should_generate_embed_from_notification(self):
        # given
        notification = Notification.objects.get(notification_id=1000000403)
        notification_embed = NotificationBaseEmbed.create(notification)
        # when
        discord_embed = notification_embed.generate_embed()
        # then
        self.assertIsInstance(discord_embed, dhooks_lite.Embed)
        self.assertEqual(discord_embed.footer.text, "Eve Online")
        self.assertIn("eve_symbol_128.png", discord_embed.footer.icon_url)

    def test_should_generate_embed_from_notification_with_custom_color(self):
        # given
        notification = Notification.objects.get(notification_id=1000000403)
        notification._color_override = Webhook.Color.SUCCESS
        notification_embed = NotificationBaseEmbed.create(notification)
        # when
        discord_embed = notification_embed.generate_embed()
        # then
        self.assertIsInstance(discord_embed, dhooks_lite.Embed)
        self.assertEqual(discord_embed.color, Webhook.Color.SUCCESS)

    def test_should_generate_embed_from_notification_without_custom_color(self):
        # given
        notification = Notification.objects.get(notification_id=1000000403)
        notification_embed = NotificationBaseEmbed.create(notification)
        # when
        discord_embed = notification_embed.generate_embed()
        # then
        self.assertIsInstance(discord_embed, dhooks_lite.Embed)
        self.assertNotEqual(discord_embed.color, Webhook.Color.SUCCESS)

    def test_should_generate_embed_from_notification_with_ping_type_override(self):
        # given
        notification = Notification.objects.get(notification_id=1000000403)
        notification._ping_type_override = Webhook.PingType.EVERYONE
        notification_embed = NotificationBaseEmbed.create(notification)
        # when
        notification_embed.generate_embed()
        # then
        self.assertEqual(notification_embed.ping_type, Webhook.PingType.EVERYONE)

    def test_should_generate_embed_for_all_supported_esi_notification_types_minimal(
        self,
    ):
        for notif_type in NotificationType.esi_notifications():
            with self.subTest(notif_type=notif_type):
                # given
                notif = (
                    Notification.objects.select_related("owner", "sender")
                    .filter(notif_type=notif_type)
                    .first()
                )
                self.assertIsInstance(notif, Notification)
                notif_embed = NotificationBaseEmbed.create(notif)

                # when
                obj = notif_embed.generate_embed()

                # then
                self.assertIsInstance(obj, dhooks_lite.Embed)

    def test_should_generate_embed_for_all_supported_esi_notification_types_normal(
        self,
    ):
        # Pre-generate structures referenced in notifications
        StructureFactory(owner=self.owner, id=1000000000001, eve_type_name="Astrahus")
        StructureFactory(owner=self.owner, id=1000000000002, eve_type_name="Athanor")

        for notif_type in NotificationType.esi_notifications():
            with self.subTest(notif_type=notif_type):
                # given
                notif = (
                    Notification.objects.select_related("owner", "sender")
                    .filter(notif_type=notif_type)
                    .first()
                )
                self.assertIsInstance(notif, Notification)
                notif_embed = NotificationBaseEmbed.create(notif)

                # when
                obj = notif_embed.generate_embed()

                # then
                self.assertIsInstance(obj, dhooks_lite.Embed)
                self.assertTrue(obj.description)

    def test_should_set_ping_everyone_for_color_danger(self):
        # given
        notification = Notification.objects.get(notification_id=1000000513)
        notification_embed = NotificationBaseEmbed.create(notification)
        notification_embed._color = Webhook.Color.DANGER
        # when
        notification_embed.generate_embed()
        # then
        self.assertEqual(notification_embed.ping_type, Webhook.PingType.EVERYONE)

    def test_should_set_ping_everyone_for_color_warning(self):
        # given
        notification = Notification.objects.get(notification_id=1000000513)
        notification_embed = NotificationBaseEmbed.create(notification)
        notification_embed._color = Webhook.Color.WARNING
        # when
        notification_embed.generate_embed()
        # then
        self.assertEqual(notification_embed.ping_type, Webhook.PingType.HERE)

    def test_should_not_set_ping_everyone_for_color_info(self):
        # given
        notification = Notification.objects.get(notification_id=1000000513)
        notification_embed = NotificationBaseEmbed.create(notification)
        notification_embed._color = Webhook.Color.INFO
        # when
        notification_embed.generate_embed()
        # then
        self.assertEqual(notification_embed.ping_type, Webhook.PingType.NONE)

    def test_should_not_set_ping_everyone_for_color_success(self):
        # given
        notification = Notification.objects.get(notification_id=1000000513)
        notification_embed = NotificationBaseEmbed.create(notification)
        notification_embed._color = Webhook.Color.SUCCESS
        # when
        notification_embed.generate_embed()
        # then
        self.assertEqual(notification_embed.ping_type, Webhook.PingType.NONE)

    @override_settings(DEBUG=True)
    def test_should_set_footer_in_developer_mode(self):
        # given
        notification = Notification.objects.get(notification_id=1000000403)
        notification_embed = NotificationBaseEmbed.create(notification)
        # when
        discord_embed = notification_embed.generate_embed()
        # then
        self.assertTrue(discord_embed.footer)

    def test_should_set_special_footer_for_generated_notifications(self):
        # given
        structure = StructureFactory(owner=self.owner)
        notification = Notification.create_from_structure(
            structure, notif_type=NotificationType.STRUCTURE_FUEL_ALERT
        )
        notification_embed = NotificationBaseEmbed.create(notification)
        # when
        discord_embed = notification_embed.generate_embed()
        # then
        self.assertEqual(discord_embed.footer.text, "Structures")
        self.assertIn("structures_logo.png", discord_embed.footer.icon_url)

    def test_should_not_break_with_too_large_description(self):
        # given
        notification = Notification.objects.get(notification_id=1000000403)
        notification_embed = NotificationBaseEmbed.create(notification)
        notification_embed._description = "x" * 2049
        # when
        discord_embed = notification_embed.generate_embed()
        # then
        self.assertIsInstance(discord_embed, dhooks_lite.Embed)

    def test_should_not_break_with_too_large_title(self):
        # given
        notification = Notification.objects.get(notification_id=1000000403)
        notification_embed = NotificationBaseEmbed.create(notification)
        notification_embed._title = "x" * 257
        # when
        discord_embed = notification_embed.generate_embed()
        # then
        self.assertIsInstance(discord_embed, dhooks_lite.Embed)


class TestNotificationEmbedsClasses(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        user = UserMainDefaultOwnerFactory()
        cls.owner = OwnerFactory(user=user)
        cls.moon_id = 40161465
        cls.solar_system_id = 30002537
        cls.type_id = 16213
        EveEntityCorporationFactory(id=1000137, name="DED")

    def test_should_generate_embed_for_normal_tower_resource_alert(self):
        # given
        StarbaseFactory(
            owner=self.owner,
            eve_moon_id=self.moon_id,
            eve_solar_system_id=self.solar_system_id,
            eve_type_id=self.type_id,
        )
        data = {
            "corpID": self.owner.corporation.corporation_id,
            "moonID": self.moon_id,
            "solarSystemID": self.solar_system_id,
            "typeID": self.type_id,
            "wants": [{"quantity": 120, "typeID": 4051}],
        }
        notification = NotificationFactory(
            owner=self.owner,
            notif_type=NotificationType.TOWER_RESOURCE_ALERT_MSG,
            text_from_dict=data,
        )
        notification_embed = NotificationBaseEmbed.create(notification)
        # when
        discord_embed = notification_embed.generate_embed()
        # then
        description = markdown_to_plain(discord_embed.description)
        self.assertIn("is running out of fuel in 3 hours", description)

    def test_should_generate_embed_for_generated_tower_resource_alert(self):
        # given
        structure = StarbaseFactory(
            owner=self.owner,
            eve_moon_id=self.moon_id,
            eve_solar_system_id=self.solar_system_id,
            eve_type_id=self.type_id,
            fuel_expires_at=now() + dt.timedelta(hours=2, seconds=20),
        )
        notification = Notification.create_from_structure(
            structure=structure, notif_type=NotificationType.TOWER_RESOURCE_ALERT_MSG
        )
        notification_embed = NotificationBaseEmbed.create(notification)
        # when
        discord_embed = notification_embed.generate_embed()
        # then
        description = markdown_to_plain(discord_embed.description)
        self.assertIn("is running out of fuel in 2 hours", description)


class TestGeneratedNotification(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

    def test_should_create_tower_reinforced_embed(self):
        # given
        notif = GeneratedNotificationFactory()
        # when
        obj = NotificationBaseEmbed.create(notif)
        # then
        self.assertIsInstance(obj, NotificationTowerReinforcedExtra)

    def test_should_generate_embed(self):
        # given
        notif = GeneratedNotificationFactory()
        embed = NotificationBaseEmbed.create(notif)
        # when
        obj = embed.generate_embed()
        # then
        self.assertIsInstance(obj, dhooks_lite.Embed)
        starbase = notif.structures.first()
        self.assertIn(starbase.name, obj.description)


class TestEveNotificationEmbeds(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        cls.owner = OwnerFactory()

    def test_should_create_sov_embed(self):
        # given
        notif = NotificationFactory(
            owner=self.owner,
            sender=EveEntityAllianceFactory(),
            notif_type=NotificationType.SOV_ENTOSIS_CAPTURE_STARTED,
            text_from_dict={"solarSystemID": 30000474, "structureTypeID": 32226},
        )
        embed = NotificationBaseEmbed.create(notif)
        # when
        obj = embed.generate_embed()
        # then
        self.assertTrue(obj.description)

    def test_should_create_sov_embed_without_sender(self):
        # given
        notif = NotificationFactory(
            owner=self.owner,
            sender=None,
            notif_type=NotificationType.SOV_ENTOSIS_CAPTURE_STARTED,
            text_from_dict={"solarSystemID": 30000474, "structureTypeID": 32226},
        )
        embed = NotificationBaseEmbed.create(notif)
        # when
        obj = embed.generate_embed()
        # then
        self.assertTrue(obj.description)



"""Main logic for generating notification embeds."""

# pylint: disable=missing-class-docstring


import dhooks_lite

from django.db import models
from django.utils.translation import gettext as _

from app_utils.datetime import ldap_time_2_datetime
from app_utils.helpers import humanize_number

from structures.constants import EveTypeId
from structures.helpers import get_or_create_eve_entity, get_or_create_eve_type
from structures.models import Notification, Webhook

from .helpers import (
    gen_eve_entity_link,
    gen_solar_system_text,
    target_datetime_formatted,
)
from .main import NotificationBaseEmbed


class BillType(models.IntegerChoices):
    """A bill type for infrastructure hub bills."""

    UNKNOWN = 0, _("Unknown Bill")
    INFRASTRUCTURE_HUB = 7, _("Infrastructure Hub Bill")

    @classmethod
    def to_enum(cls, bill_id: int):
        """Create a new enum from a bill type ID."""
        try:
            return cls(bill_id)
        except ValueError:
            return cls.UNKNOWN


class NotificationBillingBillOutOfMoneyMsg(NotificationBaseEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        bill_type_id = self._data["billTypeID"]
        bill_type_str = BillType.to_enum(bill_type_id).label
        due_date = ldap_time_2_datetime(self._data["dueDate"])
        self._title = _("Insufficient Funds for Bill")
        self._description = _(
            "The selected corporation wallet division for automatic payments "
            "does not have enough current funds available to pay the %(bill_type)s "
            "due to be paid by %(due_date)s. "
            "Transfer additional funds to the selected wallet "
            "division in order to meet your pending automatic bills."
        ) % {
            "bill_type": bill_type_str,
            "due_date": target_datetime_formatted(due_date),
        }
        self._color = Webhook.Color.WARNING


class NotificationBillingIHubBillAboutToExpire(NotificationBaseEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        solar_system_link = gen_solar_system_text(self._notification.eve_solar_system())
        due_date = ldap_time_2_datetime(self._data.get("dueDate"))
        self._title = _("IHub Bill About to Expire")
        self._description = _(
            "Maintenance bill for Infrastructure Hub in %(solar_system)s "
            "expires at %(due_date)s, "
            "if not paid in time this Infrastructure Hub will self-destruct."
        ) % {
            "solar_system": solar_system_link,
            "due_date": target_datetime_formatted(due_date),
        }
        self._color = Webhook.Color.DANGER
        structure_type = get_or_create_eve_type(id=EveTypeId.IHUB)
        self._thumbnail = dhooks_lite.Thumbnail(
            structure_type.icon_url(size=self.ICON_DEFAULT_SIZE)
        )


class NotificationBillingIHubDestroyedByBillFailure(NotificationBaseEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        solar_system_link = gen_solar_system_text(self._notification.eve_solar_system())
        structure_type = self._notification.eve_structure_type()
        self._title = (
            _("%s has self-destructed due to unpaid maintenance bills")
            % structure_type.name
        )
        self._description = _(
            "%(structure_type)s in %(solar_system)s has self-destructed, "
            "as the standard maintenance bills where not paid."
        ) % {"structure_type": structure_type.name, "solar_system": solar_system_link}
        self._color = Webhook.Color.DANGER
        self._thumbnail = dhooks_lite.Thumbnail(
            structure_type.icon_url(size=self.ICON_DEFAULT_SIZE)
        )


class NotificationCorpAllBillMsg(NotificationBaseEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        amount = self._data["amount"]
        bill_type_id = self._data["billTypeID"]
        bill_type_str = BillType.to_enum(bill_type_id).label
        current_date = ldap_time_2_datetime(self._data["currentDate"])
        due_date = ldap_time_2_datetime(self._data["dueDate"])
        creditor = get_or_create_eve_entity(id=self._data["creditorID"])
        debtor = get_or_create_eve_entity(id=self._data["debtorID"])
        self._title = _("Bill issued")
        self._description = _(
            "A bill of %(amount)s ISK, due %(due_date)s owed by %(debtor)s "
            "to %(creditor)s was issued %(current_date)s. "
            "This bill is for %(bill_type)s."
        ) % {
            "amount": humanize_number(amount),
            "bill_type": bill_type_str,
            "creditor": gen_eve_entity_link(creditor),
            "debtor": gen_eve_entity_link(debtor),
            "current_date": target_datetime_formatted(current_date),
            "due_date": target_datetime_formatted(due_date),
        }
        self._color = Webhook.Color.WARNING
        self._thumbnail = dhooks_lite.Thumbnail(
            debtor.icon_url(size=self.ICON_DEFAULT_SIZE)
        )



"""Command to update planets for all known POCOs."""

from django.core.management.base import BaseCommand
from eveuniverse.models import EveSolarSystem

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from structures import __title__
from structures.constants import EveTypeId
from structures.models import Structure

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class Command(BaseCommand):
    help = "Update planets for all known POCOs."

    def handle(self, *args, **options):
        solar_system_ids = set(
            Structure.objects.filter(eve_type_id=EveTypeId.CUSTOMS_OFFICE).values_list(
                "eve_solar_system_id", flat=True
            )
        )
        self.stdout.write(
            f"Updating planets for {len(solar_system_ids)} solar systems", ending=""
        )
        for solar_system_id in solar_system_ids:
            EveSolarSystem.objects.update_or_create_esi(
                id=solar_system_id,
                include_children=True,
                wait_for_children=True,
                enabled_sections=[EveSolarSystem.Section.PLANETS],
            )
            self.stdout.write(".", ending="")

        self.stdout.write()
        self.stdout.write(self.style.SUCCESS("DONE"))



from eveuniverse.models import EveType

from app_utils.testing import NoSocketsTestCase

from structures.core import starbases
from structures.tests.testdata.load_eveuniverse import load_eveuniverse


class TestStarbases(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        cls.type_astrahus = EveType.objects.get(id=35832)
        cls.type_poco = EveType.objects.get(id=2233)
        cls.type_starbase = EveType.objects.get(id=16213)

    def test_is_starbase(self):
        self.assertFalse(starbases.is_starbase(self.type_astrahus))
        self.assertFalse(starbases.is_starbase(self.type_poco))
        self.assertTrue(starbases.is_starbase(self.type_starbase))

    def test_starbase_fuel_consumption_per_hour(self):
        # large
        obj = EveType.objects.get(id=16213)
        self.assertEqual(starbases.fuel_per_hour(obj), 40)

        # medium
        obj = EveType.objects.get(id=20061)
        self.assertEqual(starbases.fuel_per_hour(obj), 20)

        # small
        obj = EveType.objects.get(id=20062)
        self.assertEqual(starbases.fuel_per_hour(obj), 10)

        # none
        obj = EveType.objects.get(id=35832)
        self.assertIsNone(starbases.fuel_per_hour(obj))

    def test_returns_large_for_large_control_tower(self):
        obj = EveType.objects.get(id=16213)
        expected = starbases.StarbaseSize.LARGE
        self.assertEqual(starbases.starbase_size(obj), expected)

    def test_returns_medium_for_medium_control_tower(self):
        obj = EveType.objects.get(id=20061)
        expected = starbases.StarbaseSize.MEDIUM
        self.assertEqual(starbases.starbase_size(obj), expected)

    def test_returns_small_for_small_control_tower(self):
        obj = EveType.objects.get(id=20062)
        expected = starbases.StarbaseSize.SMALL
        self.assertEqual(starbases.starbase_size(obj), expected)

    def test_returns_none_for_non_control_towers(self):
        obj = EveType.objects.get(id=35832)
        self.assertIsNone(starbases.starbase_size(obj))


class TestStarbasesFuelDuration(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

    def test_can_calculate_for_large_tower(self):
        # given
        large_tower_type = EveType.objects.get(name="Caldari Control Tower")
        # when
        result = starbases.fuel_duration(
            starbase_type=large_tower_type, fuel_quantity=80
        )
        # then
        self.assertEqual(result, 7200)

    def test_can_calculate_for_large_tower_with_sov(self):
        # given
        large_tower_type = EveType.objects.get(name="Caldari Control Tower")
        # when
        result = starbases.fuel_duration(
            starbase_type=large_tower_type, fuel_quantity=80, has_sov=True
        )
        # then
        self.assertEqual(result, 9600)

    def test_can_raise_error_when_not_starbase_type(self):
        # given
        astrahus_type = EveType.objects.get(name="Astrahus")
        # when
        with self.assertRaises(ValueError):
            starbases.fuel_duration(starbase_type=astrahus_type, fuel_quantity=80)



"""functions for loading test data and for building mocks"""

import datetime as dt
import json
import logging
import unicodedata
from collections import namedtuple
from pathlib import Path
from random import randrange

import pytz
from bs4 import BeautifulSoup
from markdown import markdown

from django.forms.models import model_to_dict
from django.utils.timezone import now
from eveuniverse.models import EveEntity

from structures.core.notification_types import NotificationType
from structures.models import Notification, Owner

from .factories import NotificationFactory

_current_folder = Path(__file__).parent
_FILENAME_EVEUNIVERSE_TESTDATA = "eveuniverse.json"

logger = logging.getLogger(__name__)


def test_data_filename():
    return f"{_current_folder}/{_FILENAME_EVEUNIVERSE_TESTDATA}"


##############################
# internal functions


def _load_testdata_entities() -> dict:
    with (_current_folder / "entities.json").open("r") as f:
        entities = json.load(f)

    # update timestamp to current
    for notification in entities["Notification"]:
        notification["timestamp"] = now() - dt.timedelta(
            hours=randrange(3), minutes=randrange(60), seconds=randrange(60)
        )

    return entities


entities_testdata = _load_testdata_entities()


###################################
# helper functions
#


def load_notification_by_type(
    owner: Owner, notif_type: NotificationType
) -> Notification:
    for notification in entities_testdata["Notification"]:
        if notification["type"] == notif_type.value:
            return NotificationFactory(
                owner=owner,
                notif_type=notif_type.value,
                text=notification.get("text", ""),
            )
    raise ValueError(f"Could not find notif for type: {notif_type}")


def load_notification_entities(owner: Owner, in_bulk=True):
    """Loads notification fixtures for this owner.

    Note that the notification require some EveEntity objects to exit,
    which can be created with ``load_eve_entities()``.

    Args:
    - in_bulk: When disabled, will load notifications one by one (for debugging)
    """
    timestamp_start = now() - dt.timedelta(hours=2)
    objs = [
        _generate_notif_obj_for_owner(owner, timestamp_start, notification)
        for notification in entities_testdata["Notification"]
    ]
    if in_bulk:
        Notification.objects.bulk_create(objs)
    else:
        for obj in objs:
            logger.info("Creating notif: %s", model_to_dict(obj))
            obj.save()


def _generate_notif_obj_for_owner(
    owner: Owner, timestamp_start: dt.datetime, notification: dict
):
    notification_id = notification["notification_id"]
    text = notification["text"] if "text" in notification else None
    is_read = notification["is_read"] if "is_read" in notification else None
    timestamp_start = timestamp_start + dt.timedelta(minutes=5)
    params = {
        "notification_id": notification_id,
        "owner": owner,
        "sender_id": notification["sender_id"],
        "timestamp": timestamp_start,
        "notif_type": notification["type"],
        "text": text,
        "is_read": is_read,
        "last_updated": now(),
        "is_sent": False,
    }
    return Notification(**params)


def markdown_to_plain(text: str) -> str:
    """Convert text in markdown to plain text."""
    html = markdown(text)
    text = "".join(BeautifulSoup(html, features="html.parser").findAll(text=True))
    return unicodedata.normalize("NFKD", text)


# def generate_eve_entities_from_auth_entities():
#     """Generate EveEntity objects from existing Auth EveOnline objects."""

#     def add_eve_entity(id, name, category):
#         if id not in existing_ids:
#             objs.append(EveEntity(id=id, category=category, name=name))
#             existing_ids.add(id)

#     existing_ids = set(EveEntity.objects.values_list("id", flat=True))
#     objs = []
#     for character in EveCharacter.objects.exclude(character_id__in=existing_ids):
#         add_eve_entity(
#             id=character.character_id,
#             name=character.character_name,
#             category=EveEntity.CATEGORY_CHARACTER,
#         )
#         add_eve_entity(
#             id=character.corporation_id,
#             name=character.corporation_name,
#             category=EveEntity.CATEGORY_CORPORATION,
#         )
#         if character.alliance_id:
#             add_eve_entity(
#                 id=character.alliance_id,
#                 name=character.alliance_name,
#                 category=EveEntity.CATEGORY_ALLIANCE,
#             )

#     for corporation in EveCorporationInfo.objects.exclude(
#         corporation_id__in=existing_ids
#     ):
#         add_eve_entity(
#             id=corporation.corporation_id,
#             name=corporation.corporation_name,
#             category=EveEntity.CATEGORY_CORPORATION,
#         )

#     for alliance in EveAllianceInfo.objects.exclude(alliance_id__in=existing_ids):
#         add_eve_entity(
#             id=alliance.alliance_id,
#             name=alliance.alliance_name,
#             category=EveEntity.CATEGORY_ALLIANCE,
#         )

#     EveEntity.objects.bulk_create(objs)


def load_eve_entities():
    """Load eve entity fixtures. Will skip already existing objs."""
    existing_ids = set(EveEntity.objects.values_list("id", flat=True))
    data = {obj["id"]: obj for obj in entities_testdata["EveEntity"]}
    incoming_ids = set(data.keys())
    missing_ids = incoming_ids - existing_ids
    objs = [EveEntity(**data[entity_id]) for entity_id in missing_ids]
    if objs:
        EveEntity.objects.bulk_create(objs)
    logger.info("Loaded %d EveEntity objects", len(objs))
    return objs


def clone_notification(obj: Notification) -> Notification:
    """Return clone of a Notification."""
    new_object = NotificationFactory(
        sender=obj.sender, notif_type=obj.notif_type, owner=obj.owner, text=obj.text
    )
    return new_object


def datetime_to_ldap(my_dt: dt.datetime) -> int:
    """Return a standard datetime as ldap datetime."""
    return (
        ((my_dt - dt.datetime(1970, 1, 1, tzinfo=pytz.utc)).total_seconds())
        + 11644473600
    ) * 10000000


NearestCelestial = namedtuple(
    "NearestCelestial", ["eve_type", "eve_object", "distance"]
)



# Generated by Django 4.0.10 on 2023-10-30 17:26

import django.core.validators
import django.db.models.deletion
import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("eveuniverse", "0010_alter_eveindustryactivityduration_eve_type_and_more"),
        ("authentication", "0021_alter_userprofile_language"),
        ("structures", "0003_add_localization_and_unique_key"),
    ]

    operations = [
        migrations.AlterField(
            model_name="fuelalert",
            name="hours",
            field=models.PositiveIntegerField(
                db_index=True,
                help_text="Number of hours before fuel expiration this alert was sent",
                verbose_name="hours",
            ),
        ),
        migrations.AlterField(
            model_name="generatednotification",
            name="notif_type",
            field=models.CharField(
                db_index=True,
                default="",
                help_text="Type of this notification",
                max_length=100,
                verbose_name="type",
            ),
        ),
        migrations.AlterField(
            model_name="notification",
            name="notif_type",
            field=models.CharField(
                db_index=True,
                default="",
                help_text="Type of this notification",
                max_length=100,
                verbose_name="type",
            ),
        ),
        migrations.AlterField(
            model_name="owner",
            name="forwarding_last_update_at",
            field=models.DateTimeField(
                blank=True,
                default=None,
                help_text="When the last successful update happened",
                null=True,
                verbose_name="forwarding last update at",
            ),
        ),
        migrations.AlterField(
            model_name="owner",
            name="has_default_pings_enabled",
            field=models.BooleanField(
                default=True,
                help_text="To enable or disable pinging of notifications for this owner e.g. with @everyone and @here",
                verbose_name="has default pings enabled",
            ),
        ),
        migrations.AlterField(
            model_name="owner",
            name="is_active",
            field=models.BooleanField(
                default=True,
                help_text="Whether this owner is currently included in the sync process",
                verbose_name="is active",
            ),
        ),
        migrations.AlterField(
            model_name="owner",
            name="is_alliance_main",
            field=models.BooleanField(
                default=False,
                help_text="Whether alliance wide notifications are forwarded for this owner (e.g. sov notifications)",
                verbose_name="is alliance main",
            ),
        ),
        migrations.AlterField(
            model_name="owner",
            name="is_included_in_service_status",
            field=models.BooleanField(
                default=True,
                help_text="Whether the sync status of this owner is included in the overall status of this services",
                verbose_name="is included in service status",
            ),
        ),
        migrations.AlterField(
            model_name="owner",
            name="is_up",
            field=models.BooleanField(
                default=None,
                editable=False,
                help_text="Whether all services for this owner are currently up",
                null=True,
                verbose_name="is up",
            ),
        ),
        migrations.AlterField(
            model_name="owner",
            name="notifications_last_update_at",
            field=models.DateTimeField(
                blank=True,
                default=None,
                help_text="When the last successful update happened",
                null=True,
                verbose_name="notifications last update at",
            ),
        ),
        migrations.AlterField(
            model_name="owner",
            name="structures_last_update_at",
            field=models.DateTimeField(
                blank=True,
                default=None,
                help_text="When the last successful update happened",
                null=True,
                verbose_name="structures last update at",
            ),
        ),
        migrations.AlterField(
            model_name="ownercharacter",
            name="character_ownership",
            field=models.ForeignKey(
                help_text="Character used for syncing",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="authentication.characterownership",
                verbose_name="character_ownership",
            ),
        ),
        migrations.AlterField(
            model_name="ownercharacter",
            name="notifications_last_used_at",
            field=models.DateTimeField(
                db_index=True,
                default=None,
                editable=False,
                help_text="When this character was last used for syncing notifications",
                null=True,
                verbose_name="notifications last used at",
            ),
        ),
        migrations.AlterField(
            model_name="ownercharacter",
            name="structures_last_used_at",
            field=models.DateTimeField(
                db_index=True,
                default=None,
                editable=False,
                help_text="When this character was last used for syncing structures",
                null=True,
                verbose_name="structures last used at",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="created_at",
            field=models.DateTimeField(
                default=django.utils.timezone.now,
                help_text="Date this structure was received from ESI for the first time",
                verbose_name="created at",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="has_core",
            field=models.BooleanField(
                blank=True,
                db_index=True,
                default=None,
                help_text="Whether the structure has a quantum core",
                null=True,
                verbose_name="has core",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="has_fitting",
            field=models.BooleanField(
                blank=True,
                db_index=True,
                default=None,
                help_text="Whether the structure has a fitting",
                null=True,
                verbose_name="has fitting",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="last_online_at",
            field=models.DateTimeField(
                blank=True,
                default=None,
                help_text="Date this structure had any of it's services online",
                null=True,
                verbose_name="last online at",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="last_updated_at",
            field=models.DateTimeField(
                blank=True,
                default=None,
                help_text="Date this structure was last updated from the EVE server",
                null=True,
                verbose_name="last updated at",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="position_x",
            field=models.FloatField(
                blank=True,
                default=None,
                help_text="X coordinate of position in the solar system",
                null=True,
                verbose_name="position x",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="position_y",
            field=models.FloatField(
                blank=True,
                default=None,
                help_text="Y coordinate of position in the solar system",
                null=True,
                verbose_name="position y",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="position_z",
            field=models.FloatField(
                blank=True,
                default=None,
                help_text="Z coordinate of position in the solar system",
                null=True,
                verbose_name="position z",
            ),
        ),
        migrations.AlterField(
            model_name="structureitem",
            name="eve_type",
            field=models.ForeignKey(
                help_text="Type of the item",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="eveuniverse.evetype",
                verbose_name="type",
            ),
        ),
        migrations.AlterField(
            model_name="structuretag",
            name="description",
            field=models.TextField(
                blank=True,
                default=None,
                help_text="Description for this tag",
                null=True,
                verbose_name="description",
            ),
        ),
        migrations.AlterField(
            model_name="structuretag",
            name="is_default",
            field=models.BooleanField(
                default=False,
                help_text="When enabled this custom tag will automatically be added to new structures",
                verbose_name="is default",
            ),
        ),
        migrations.AlterField(
            model_name="structuretag",
            name="is_user_managed",
            field=models.BooleanField(
                default=True,
                help_text="When disabled this tag is created and managed by the system and can not be modified by users",
                verbose_name="is user managed",
            ),
        ),
        migrations.AlterField(
            model_name="structuretag",
            name="name",
            field=models.CharField(
                help_text="Name of the tag, which must be unique",
                max_length=255,
                unique=True,
                verbose_name="name",
            ),
        ),
        migrations.AlterField(
            model_name="structuretag",
            name="order",
            field=models.PositiveIntegerField(
                blank=True,
                default=100,
                help_text="Number defining the order tags are shown. custom tags can not have an order below 100",
                validators=[django.core.validators.MinValueValidator(100)],
                verbose_name="order",
            ),
        ),
        migrations.AlterField(
            model_name="structuretag",
            name="style",
            field=models.CharField(
                blank=True,
                choices=[
                    ("default", "grey"),
                    ("primary", "dark blue"),
                    ("success", "green"),
                    ("info", "light blue"),
                    ("warning", "orange"),
                    ("danger", "red"),
                ],
                default="default",
                help_text="Color style of tag",
                max_length=16,
                verbose_name="style",
            ),
        ),
        migrations.AlterField(
            model_name="webhook",
            name="is_active",
            field=models.BooleanField(
                default=True,
                help_text="Whether notifications are currently sent to this webhook",
                verbose_name="is active",
            ),
        ),
        migrations.AlterField(
            model_name="webhook",
            name="language_code",
            field=models.CharField(
                blank=True,
                choices=[
                    ("en", "English"),
                    ("de", "German"),
                    ("es", "Spanish"),
                    ("zh-hans", "Chinese Simplified"),
                    ("ru", "Russian"),
                    ("ko", "Korean"),
                ],
                default=None,
                help_text="Language of notifications send to this webhook",
                max_length=8,
                null=True,
                verbose_name="language",
            ),
        ),
        migrations.AlterField(
            model_name="webhook",
            name="name",
            field=models.CharField(
                help_text="Short name to identify this webhook",
                max_length=64,
                unique=True,
                verbose_name="name",
            ),
        ),
        migrations.AlterField(
            model_name="webhook",
            name="notes",
            field=models.TextField(
                blank=True,
                default=None,
                help_text="Notes regarding this webhook",
                null=True,
                verbose_name="notes",
            ),
        ),
        migrations.AlterField(
            model_name="webhook",
            name="url",
            field=models.CharField(
                help_text="URL of this webhook, e.g. https://discordapp.com/api/webhooks/123456/abcdef",
                max_length=255,
                unique=True,
                verbose_name="url",
            ),
        ),
        migrations.AlterField(
            model_name="webhook",
            name="webhook_type",
            field=models.IntegerField(
                choices=[(1, "Discord Webhook")],
                default=1,
                help_text="Type of this webhook",
                verbose_name="webhook type",
            ),
        ),
    ]



import datetime as dt
from copy import deepcopy
from unittest.mock import patch

from pytz import UTC

from django.utils.timezone import now

from app_utils.testing import NoSocketsTestCase

from structures.constants import EveCorporationId, EveTypeId
from structures.core.notification_types import NotificationType
from structures.models import (
    EveSpaceType,
    JumpFuelAlertConfig,
    PocoDetails,
    Structure,
    StructureItem,
    StructureService,
    StructureTag,
)
from structures.tests.testdata.factories import (
    EveCharacterFactory,
    EveCorporationInfoFactory,
    EveEntityCorporationFactory,
    EveSovereigntyMapFactory,
    FuelAlertConfigFactory,
    JumpGateFactory,
    OwnerFactory,
    PocoDetailsFactory,
    PocoFactory,
    SkyhookFactory,
    StarbaseFactory,
    StructureFactory,
    StructureItemFactory,
    StructureServiceFactory,
    StructureTagFactory,
)
from structures.tests.testdata.load_eveuniverse import load_eveuniverse

STRUCTURES_PATH = "structures.models.structures_1"
NOTIFICATIONS_PATH = "structures.models.notifications"

EVE_ID_HELIUM_FUEL_BLOCK = 4247
EVE_ID_NITROGEN_FUEL_BLOCK = 4051


class TestPocoDetails(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        cls.owner = OwnerFactory()
        cls.structure = PocoFactory(owner=cls.owner, poco_details=False)

    def test_should_return_tax_and_access_for_corporation_member(self):
        # given
        details = PocoDetailsFactory(
            structure=self.structure, corporation_tax_rate=0.01
        )
        character = EveCharacterFactory(corporation=details.structure.owner.corporation)
        # when
        result = details.determine_access_and_tax_for_character(character)
        # then
        self.assertTrue(result.has_access)
        self.assertTrue(result.is_confident)
        self.assertEqual(result.tax_rate, 0.01)

    def test_should_return_tax_and_access_for_alliance_member(self):
        # given
        details = PocoDetailsFactory(
            structure=self.structure, allow_alliance_access=True, alliance_tax_rate=0.02
        )
        corporation = EveCorporationInfoFactory(
            alliance=self.owner.corporation.alliance
        )
        character = EveCharacterFactory(corporation=corporation)
        # when
        result = details.determine_access_and_tax_for_character(character)
        # then
        self.assertTrue(result.has_access)
        self.assertTrue(result.is_confident)
        self.assertEqual(result.tax_rate, 0.02)

    def test_should_return_tax_for_random_1_when_allowed(self):
        # given
        details = PocoDetailsFactory(
            structure=self.structure,
            allow_access_with_standings=True,
            standing_level=PocoDetails.StandingLevel.NEUTRAL,
            neutral_standing_tax_rate=0.04,
        )
        character = EveCharacterFactory()
        # when
        result = details.determine_access_and_tax_for_character(character)
        # then
        self.assertTrue(result.has_access)
        self.assertFalse(result.is_confident)
        self.assertEqual(result.tax_rate, 0.04)

    def test_should_return_tax_for_random_2_when_allowed(self):
        # given
        details = PocoDetailsFactory(
            structure=self.structure,
            allow_access_with_standings=True,
            standing_level=PocoDetails.StandingLevel.NEUTRAL,
            neutral_standing_tax_rate=0,
        )
        character = EveCharacterFactory()
        # when
        result = details.determine_access_and_tax_for_character(character)
        # then
        self.assertTrue(result.has_access)
        self.assertFalse(result.is_confident)
        self.assertEqual(result.tax_rate, 0)

    def test_should_return_tax_for_random_3_when_allowed(self):
        # given
        details = PocoDetailsFactory(
            structure=self.structure,
            allow_access_with_standings=True,
            standing_level=PocoDetails.StandingLevel.TERRIBLE,
            neutral_standing_tax_rate=0.05,
        )
        character = EveCharacterFactory()
        # when
        result = details.determine_access_and_tax_for_character(character)
        # then
        self.assertTrue(result.has_access)
        self.assertFalse(result.is_confident)
        self.assertEqual(result.tax_rate, 0.05)

    def test_should_return_no_access_for_random(self):
        # given
        details = PocoDetailsFactory(
            structure=self.structure, allow_access_with_standings=False
        )
        character = EveCharacterFactory()
        # when
        result = details.determine_access_and_tax_for_character(character)
        # then
        self.assertFalse(result.has_access)
        self.assertFalse(result.is_confident)
        self.assertIsNone(result.tax_rate)

    def test_should_return_standing_map_for_neutral_1(self):
        # given
        details = PocoDetailsFactory(structure=self.structure)
        details.standing_level = PocoDetails.StandingLevel.NEUTRAL
        details.allow_access_with_standings = True
        # when
        result = details.standing_level_access_map()
        # then
        self.assertDictEqual(
            result,
            {
                "NONE": False,
                "TERRIBLE": False,
                "BAD": False,
                "NEUTRAL": True,
                "GOOD": True,
                "EXCELLENT": True,
            },
        )

    def test_should_return_standing_map_for_neutral_2(self):
        # given
        details = PocoDetailsFactory(structure=self.structure)
        details.standing_level = PocoDetails.StandingLevel.NEUTRAL
        details.allow_access_with_standings = False
        # when
        result = details.standing_level_access_map()
        # then
        self.assertDictEqual(
            result,
            {
                "NONE": False,
                "TERRIBLE": False,
                "BAD": False,
                "NEUTRAL": False,
                "GOOD": False,
                "EXCELLENT": False,
            },
        )

    def test_should_return_standing_map_for_terrible(self):
        # given
        details = PocoDetailsFactory(structure=self.structure)
        details.standing_level = PocoDetails.StandingLevel.TERRIBLE
        details.allow_access_with_standings = True
        # when
        result = details.standing_level_access_map()
        # then
        self.assertDictEqual(
            result,
            {
                "NONE": False,
                "TERRIBLE": True,
                "BAD": True,
                "NEUTRAL": True,
                "GOOD": True,
                "EXCELLENT": True,
            },
        )


class TestStructure(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        cls.owner = OwnerFactory()

    def test_str_upwell(self):
        obj = StructureFactory.build(
            owner=self.owner,
            eve_solar_system_name="Amamake",
            name="Test Structure Alpha",
        )
        expected = "Amamake - Test Structure Alpha"
        self.assertEqual(str(obj), expected)

    def test_str_upwell_wo_solar_system(self):
        obj = StructureFactory.build(
            owner=self.owner,
            eve_solar_system=None,
            name="Test Structure Alpha",
        )
        expected = "? - Test Structure Alpha"
        self.assertEqual(str(obj), expected)

    def test_str_poco(self):
        obj = PocoFactory.build(eve_planet_name="Amamake V")
        expected = "Amamake V - Customs Office (Amamake V)"
        self.assertEqual(str(obj), expected)

    def test_str_starbase(self):
        obj = StarbaseFactory(
            eve_moon_name="Amamake II - Moon 1", name="Home Sweat Home"
        )
        expected = "Amamake II - Moon 1 - Home Sweat Home"
        self.assertEqual(str(obj), expected)

    def test_repr(self):
        obj = StructureFactory.build()
        self.assertTrue(repr(obj))

    def test_structure_is_full_power(self):
        structure = StructureFactory.build(owner=self.owner)

        # true when upwell structure and has fuel that is not expired
        structure.fuel_expires_at = now() + dt.timedelta(hours=1)
        self.assertTrue(structure.is_full_power)

        # false when upwell structure and has fuel, but is expired
        structure.fuel_expires_at = now() - dt.timedelta(hours=1)
        self.assertFalse(structure.is_full_power)

        # False when no fuel info
        structure.fuel_expires_at = None
        self.assertFalse(structure.is_full_power)

        # none when no upwell structure
        poco = PocoFactory(owner=self.owner)
        poco.fuel_expires_at = now() + dt.timedelta(hours=1)
        self.assertIsNone(poco.is_full_power)

    def test_is_low_power(self):
        structure = StructureFactory.build(owner=self.owner)

        # true if Upwell structure and fuel expired and last online < 7d
        structure.fuel_expires_at = now() - dt.timedelta(seconds=3)
        structure.last_online_at = now() - dt.timedelta(days=3)
        self.assertTrue(structure.is_low_power)

        # True if Upwell structure and no fuel info and last online < 7d
        structure.fuel_expires_at = None
        structure.last_online_at = now() - dt.timedelta(days=3)
        self.assertTrue(structure.is_low_power)

        # false if Upwell structure and it has fuel
        structure.fuel_expires_at = now() + dt.timedelta(days=3)
        self.assertFalse(structure.is_low_power)

        # none if upwell structure, but not online info
        structure.fuel_expires_at = now() - dt.timedelta(seconds=3)
        structure.last_online_at = None
        self.assertFalse(structure.is_low_power)

        structure.fuel_expires_at = None
        structure.last_online_at = None
        self.assertFalse(structure.is_low_power)

        # none for non structures
        poco = PocoFactory.build(owner=self.owner)
        self.assertIsNone(poco.is_low_power)

        starbase = StarbaseFactory.build(owner=self.owner)
        self.assertIsNone(starbase.is_low_power)

    def test_is_abandoned(self):
        structure = StructureFactory.build(owner=self.owner)

        # true when upwell structure, online > 7 days
        structure.last_online_at = now() - dt.timedelta(days=7, seconds=1)

        # false when upwell structure, online <= 7 days or none
        structure.last_online_at = now() - dt.timedelta(days=7, seconds=0)
        self.assertFalse(structure.is_abandoned)

        structure.last_online_at = now() - dt.timedelta(days=3)
        self.assertFalse(structure.is_abandoned)

        # none if missing information
        structure.last_online_at = None
        self.assertFalse(structure.is_abandoned)

        # none for non structures
        starbase = StarbaseFactory(owner=self.owner)
        self.assertIsNone(starbase.is_abandoned)

    def test_extract_name_from_esi_response(self):
        expected = "Alpha"
        self.assertEqual(
            Structure.extract_name_from_esi_response("Super - Alpha"), expected
        )
        self.assertEqual(Structure.extract_name_from_esi_response("Alpha"), expected)

    def test_should_return_hours_when_fuel_expires(self):
        # given
        obj = StructureFactory.build(owner=self.owner)
        obj.fuel_expires_at = now() + dt.timedelta(hours=2)
        # when
        result = obj.hours_fuel_expires
        # then
        self.assertAlmostEqual(result, 2.0, delta=0.1)

    def test_should_return_none_when_no_fuel_info(self):
        # given
        obj = StructureFactory.build(owner=self.owner)
        obj.fuel_expires_at = None
        # when
        result = obj.hours_fuel_expires
        # then
        self.assertIsNone(result)

    def test_should_return_moon_location(self):
        # given
        obj = StarbaseFactory.build(eve_moon_name="Amamake II - Moon 1")
        # when/then
        self.assertEqual(obj.location_name, "Amamake II - Moon 1")

    def test_should_return_planet_location(self):
        # given
        obj = PocoFactory.build(eve_planet_name="Amamake V")
        # when/then
        self.assertEqual(obj.location_name, "Amamake V")

    def test_should_return_solar_system_location(self):
        # given
        obj = StructureFactory.build(eve_solar_system_name="Amamake")
        # when/then
        self.assertEqual(obj.location_name, "Amamake")

    def test_should_return_unknown_location(self):
        # given
        obj = StructureFactory.build(eve_solar_system=None)
        # when/then
        self.assertEqual(obj.location_name, "?")

    # TODO: activate
    # def test_is_upwell_structure_data_error(self):
    #     # group without a category
    #     my_group = EveGroup.objects.create(id=299999, name="invalid group")
    #     my_type = EveType.objects.create(
    #         id=199999, name="invalid type", eve_group=my_group
    #     )
    #     self.assertFalse(my_type.is_upwell_structure)

    def test_should_detect_structure_has_position(self):
        # given
        structure = StructureFactory.build(owner=self.owner)
        # then
        self.assertTrue(structure.has_position)

    def test_should_detect_structure_has_no_position(self):
        # given
        structure = StructureFactory.build(
            position_x=None, position_y=None, position_z=None
        )
        # then
        self.assertFalse(structure.has_position)

    def test_should_not_show_structure_as_reinforced(self):
        # given
        structure = StructureFactory(
            owner=self.owner, state=Structure.State.SHIELD_VULNERABLE
        )
        # when/then
        self.assertFalse(structure.is_reinforced)

    def test_should_show_structure_as_reinforced(self):
        structure = StructureFactory.build(owner=self.owner)
        for state, excepted in [
            (Structure.State.ANCHOR_VULNERABLE, False),
            (Structure.State.ANCHORING, False),
            (Structure.State.ARMOR_REINFORCE, True),
            (Structure.State.ARMOR_VULNERABLE, False),
            (Structure.State.DEPLOY_VULNERABLE, False),
            (Structure.State.FITTING_INVULNERABLE, False),
            (Structure.State.HULL_REINFORCE, True),
            (Structure.State.HULL_VULNERABLE, False),
            (Structure.State.ONLINING_VULNERABLE, False),
            (Structure.State.SHIELD_VULNERABLE, False),
            (Structure.State.UNANCHORED, False),
            (Structure.State.NA, False),
            (Structure.State.UNKNOWN, False),
        ]:
            with self.subTest(state=state):
                structure.state = state
                self.assertIs(structure.is_reinforced, excepted)

    def test_should_show_starbase_as_reinforced(self):
        structure = StarbaseFactory.build(owner=self.owner)
        for state, excepted in [
            (Structure.State.POS_OFFLINE, False),
            (Structure.State.POS_ONLINE, False),
            (Structure.State.POS_ONLINING, False),
            (Structure.State.POS_REINFORCED, True),
            (Structure.State.POS_UNANCHORING, False),
        ]:
            with self.subTest(state=state):
                structure.state = state
                self.assertIs(structure.is_reinforced, excepted)

    def test_owner_alliance_has_sov_in_null_sec_system(self):
        obj = StructureFactory(owner=self.owner, eve_solar_system_name="1-PGSG")
        EveSovereigntyMapFactory(
            eve_solar_system_name="1-PGSG", corporation=self.owner.corporation
        )
        self.assertTrue(obj.owner_has_sov())

    def test_owner_has_no_sov_in_null_sec_system(self):
        obj = StructureFactory(owner=self.owner, eve_solar_system_name="A-C5TC")
        self.assertFalse(obj.owner_has_sov())

    def test_owner_has_no_sov_in_low_sec_system(self):
        obj = StructureFactory(owner=self.owner, eve_solar_system_name="Amamake")
        self.assertFalse(obj.owner_has_sov())


class TestStructureIsX(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        cls.owner = OwnerFactory()
        cls.jump_gate = JumpGateFactory.build(owner=cls.owner)
        cls.poco = PocoFactory.build(owner=cls.owner)
        cls.skyhook = SkyhookFactory.build(owner=cls.owner)
        cls.starbase = StarbaseFactory.build(owner=cls.owner)
        cls.upwell_structure = StructureFactory.build(owner=cls.owner)

    def test_is_jump_gate(self):
        self.assertFalse(self.upwell_structure.is_jump_gate)
        self.assertFalse(self.poco.is_jump_gate)
        self.assertFalse(self.starbase.is_jump_gate)
        self.assertTrue(self.jump_gate.is_jump_gate)
        self.assertFalse(self.skyhook.is_jump_gate)

    def test_is_poco(self):
        self.assertFalse(self.upwell_structure.is_poco)
        self.assertTrue(self.poco.is_poco)
        self.assertFalse(self.starbase.is_poco)
        self.assertFalse(self.jump_gate.is_poco)
        self.assertFalse(self.skyhook.is_poco)

    def test_is_starbase(self):
        self.assertFalse(self.upwell_structure.is_starbase)
        self.assertFalse(self.poco.is_starbase)
        self.assertTrue(self.starbase.is_starbase)
        self.assertFalse(self.jump_gate.is_starbase)
        self.assertFalse(self.skyhook.is_starbase)

    def test_is_skyhook(self):
        self.assertFalse(self.upwell_structure.is_skyhook)
        self.assertFalse(self.poco.is_skyhook)
        self.assertFalse(self.starbase.is_skyhook)
        self.assertFalse(self.jump_gate.is_skyhook)
        self.assertTrue(self.skyhook.is_skyhook)

    def test_is_upwell_structure(self):
        self.assertTrue(self.upwell_structure.is_upwell_structure)
        self.assertFalse(self.poco.is_upwell_structure)
        self.assertFalse(self.starbase.is_upwell_structure)
        self.assertTrue(self.jump_gate.is_upwell_structure)
        self.assertFalse(self.skyhook.is_upwell_structure)


class TestStructureFuel(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        cls.owner = OwnerFactory()

    def test_should_return_jump_fuel_quantity(self):
        # given
        structure = JumpGateFactory(owner=self.owner, jump_fuel_quantity=False)
        StructureItemFactory(
            structure=structure,
            eve_type_id=EveTypeId.LIQUID_OZONE,
            location_flag=StructureItem.LocationFlag.STRUCTURE_FUEL,
            quantity=32,
        )
        StructureItemFactory(
            structure=structure,
            eve_type_id=EveTypeId.LIQUID_OZONE,
            location_flag=StructureItem.LocationFlag.STRUCTURE_FUEL,
            quantity=10,
        )
        # when
        result = structure.jump_fuel_quantity()
        # then
        self.assertEqual(result, 42)

    def test_should_return_none_when_not_jump_gate_1(self):
        # given
        structure = StarbaseFactory(owner=self.owner, eve_type_name="Astrahus")
        # when
        result = structure.jump_fuel_quantity()
        # then
        self.assertIsNone(result)

    def test_should_return_none_when_not_jump_gate_2(self):
        # given
        structure = StarbaseFactory(owner=self.owner)
        # when
        result = structure.jump_fuel_quantity()
        # then
        self.assertIsNone(result)

    def test_should_return_none_when_not_jump_gate_(self):
        # given
        structure = StarbaseFactory(owner=self.owner)
        # when
        result = structure.jump_fuel_quantity()
        # then
        self.assertIsNone(result)

    def test_should_remove_fuel_alerts_when_fuel_level_above_threshold(self):
        # given
        config = JumpFuelAlertConfig.objects.create(threshold=100)
        structure = JumpGateFactory(owner=self.owner, jump_fuel_quantity=101)
        structure.jump_fuel_alerts.create(config=config)
        # when
        structure.reevaluate_jump_fuel_alerts()
        # then
        self.assertEqual(structure.jump_fuel_alerts.count(), 0)

    def test_should_keep_fuel_alerts_when_fuel_level_below_threshold(self):
        # given
        config = JumpFuelAlertConfig.objects.create(threshold=100)
        structure = JumpGateFactory(owner=self.owner, jump_fuel_quantity=99)
        structure.jump_fuel_alerts.create(config=config)
        # when
        structure.reevaluate_jump_fuel_alerts()
        # then
        self.assertEqual(structure.jump_fuel_alerts.count(), 1)

    def test_should_return_fuel_blocks(self):
        # given
        structure = StructureFactory(owner=self.owner)
        StructureItemFactory(
            structure=structure,
            eve_type_id=EVE_ID_NITROGEN_FUEL_BLOCK,
            location_flag=StructureItem.LocationFlag.STRUCTURE_FUEL,
            quantity=250,
        )
        StructureItemFactory(
            structure=structure,
            eve_type_id=EVE_ID_HELIUM_FUEL_BLOCK,
            location_flag=StructureItem.LocationFlag.STRUCTURE_FUEL,
            quantity=1000,
        )
        StructureItemFactory(
            structure=structure,
            eve_type_id=EVE_ID_HELIUM_FUEL_BLOCK,
            location_flag=StructureItem.LocationFlag.CARGO,
            quantity=500,
        )
        # when
        result = structure.structure_fuel_quantity
        # then
        self.assertEqual(result, 1250)

    def test_should_return_fuel_need(self):
        # given
        structure = StructureFactory(
            owner=self.owner,
            eve_type_name="Astrahus",
            fuel_expires_at=dt.datetime(2022, 1, 24, 5, 0, tzinfo=UTC),
        )
        with patch("django.utils.timezone.now") as mock_now:
            mock_now.return_value = dt.datetime(2021, 12, 17, 15, 13, tzinfo=UTC)
            StructureItemFactory(
                structure=structure,
                eve_type_id=EVE_ID_NITROGEN_FUEL_BLOCK,
                location_flag=StructureItem.LocationFlag.STRUCTURE_FUEL,
                quantity=6309,
            )
        # when
        result = structure.structure_fuel_usage()
        # then
        self.assertEqual(result, 168)

    def test_should_not_break_when_no_fuel_date(self):
        # given
        structure = StructureFactory(
            owner=self.owner,
            eve_type_name="Astrahus",
            fuel_expires_at=None,
        )
        with patch("django.utils.timezone.now") as mock_now:
            mock_now.return_value = dt.datetime(2021, 12, 17, 15, 13, tzinfo=UTC)
            StructureItemFactory(
                structure=structure,
                eve_type_id=EVE_ID_NITROGEN_FUEL_BLOCK,
                location_flag=StructureItem.LocationFlag.STRUCTURE_FUEL,
                quantity=6309,
            )
        # when
        result = structure.structure_fuel_usage()
        # then
        self.assertIsNone(result)

    def test_should_handle_zero_hours_remaining(self):
        # given
        my_now = now()
        structure = StructureFactory(
            owner=self.owner,
            eve_type_name="Astrahus",
            fuel_expires_at=my_now,
        )
        with patch("django.utils.timezone.now") as mock_now:
            mock_now.return_value = my_now
            StructureItemFactory(
                structure=structure,
                eve_type_id=EVE_ID_NITROGEN_FUEL_BLOCK,
                location_flag=StructureItem.LocationFlag.STRUCTURE_FUEL,
                quantity=1,
                last_updated_at=my_now,
            )
        # when
        result = structure.structure_fuel_usage()
        # then
        self.assertIsNone(result)

    def test_should_return_true_for_structure(self):
        # given
        structure = StructureFactory.build(owner=self.owner)
        # when/then
        self.assertTrue(structure.is_burning_fuel)

    def test_should_return_false_for_structure(self):
        # given
        structure = StructureFactory.build(owner=self.owner, fuel_expires_at=None)
        # when/then
        self.assertFalse(structure.is_burning_fuel)

    def test_should_return_whether_starbase_is_burning_fuel(self):
        starbase = StarbaseFactory.build(owner=self.owner)
        for state, expected in [
            (Structure.State.POS_ONLINE, True),
            (Structure.State.POS_REINFORCED, True),
            (Structure.State.POS_UNANCHORING, True),
            (Structure.State.POS_OFFLINE, False),
            (Structure.State.POS_ONLINING, False),
        ]:
            with self.subTest(state=state):
                starbase.state = state
                self.assertIs(starbase.is_burning_fuel, expected)

    def test_should_return_false_for_poco(self):
        # given
        poco = PocoFactory.build(owner=self.owner)
        # when/then
        self.assertFalse(poco.is_burning_fuel)


@patch(STRUCTURES_PATH + ".STRUCTURES_FEATURE_REFUELED_NOTIFICATIONS", True)
@patch(STRUCTURES_PATH + ".Structure.FUEL_DATES_EQUAL_THRESHOLD_UPWELL", 900)
@patch(STRUCTURES_PATH + ".Structure.FUEL_DATES_EQUAL_THRESHOLD_STARBASE", 7200)
class TestStructureFuelLevels(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        cls.owner = OwnerFactory(
            webhooks__notification_types=[NotificationType.STRUCTURE_REFUELED_EXTRA]
        )
        EveEntityCorporationFactory(
            id=EveCorporationId.DED, name="DED"
        )  # for notifications

    @patch(
        NOTIFICATIONS_PATH + ".Notification.send_to_configured_webhooks",
        lambda *args, **kwargs: None,
    )
    def test_should_reset_fuel_notifications_when_refueled_1(self):
        # given
        config = FuelAlertConfigFactory(start=48, end=0, repeat=12)
        structure = StructureFactory(
            owner=self.owner, fuel_expires_at=now() + dt.timedelta(hours=12)
        )
        structure.structure_fuel_alerts.create(config=config, hours=12)
        old_instance = deepcopy(structure)
        # when
        structure.fuel_expires_at = now() + dt.timedelta(hours=13)
        structure.handle_fuel_notifications(old_instance)
        # then
        self.assertEqual(structure.structure_fuel_alerts.count(), 0)

    @patch(
        NOTIFICATIONS_PATH + ".Notification.send_to_configured_webhooks",
        lambda *args, **kwargs: None,
    )
    def test_should_reset_fuel_notifications_when_fuel_expires_date_has_changed(self):
        # given
        config = FuelAlertConfigFactory(start=48, end=0, repeat=12)
        structure = StructureFactory(
            owner=self.owner, fuel_expires_at=now() + dt.timedelta(hours=12)
        )
        old_instance = deepcopy(structure)
        structure.structure_fuel_alerts.create(config=config, hours=12)
        # when
        structure.fuel_expires_at = now() + dt.timedelta(hours=11)
        structure.handle_fuel_notifications(old_instance)
        # then
        self.assertEqual(structure.structure_fuel_alerts.count(), 0)

    @patch(
        NOTIFICATIONS_PATH + ".Notification.send_to_configured_webhooks",
        lambda *args, **kwargs: None,
    )
    def test_should_not_reset_fuel_notifications_when_fuel_expiry_dates_unchanged(self):
        # given
        config = FuelAlertConfigFactory(start=48, end=0, repeat=12)
        structure = StructureFactory(
            owner=self.owner, fuel_expires_at=now() + dt.timedelta(hours=12)
        )
        structure.structure_fuel_alerts.create(config=config, hours=12)
        old_instance = deepcopy(structure)
        # when
        structure.fuel_expires_at = now() + dt.timedelta(hours=12, minutes=5)
        structure.handle_fuel_notifications(old_instance)
        # then
        self.assertEqual(structure.structure_fuel_alerts.count(), 1)

    @patch(NOTIFICATIONS_PATH + ".Notification.create_from_structure")
    def test_should_generate_structure_refueled_notif_when_fuel_level_increased(
        self, mock_create_from_structure
    ):
        # given
        structure = StructureFactory(
            owner=self.owner, fuel_expires_at=now() + dt.timedelta(hours=1)
        )
        old_instance = deepcopy(structure)
        # when
        structure.fuel_expires_at = now() + dt.timedelta(hours=6)
        structure.handle_fuel_notifications(old_instance)
        # then
        self.assertTrue(mock_create_from_structure.called)
        _, kwargs = mock_create_from_structure.call_args
        self.assertEqual(
            kwargs["notif_type"], NotificationType.STRUCTURE_REFUELED_EXTRA
        )

    @patch(NOTIFICATIONS_PATH + ".Notification.create_from_structure")
    def test_should_generate_tower_refueled_notif_when_fuel_level_increased(
        self, mock_create_from_structure
    ):
        # given
        structure = StarbaseFactory(
            owner=self.owner, fuel_expires_at=now() + dt.timedelta(hours=1)
        )
        old_instance = deepcopy(structure)
        # when
        structure.fuel_expires_at = now() + dt.timedelta(hours=4)
        structure.handle_fuel_notifications(old_instance)
        # then
        self.assertTrue(mock_create_from_structure.called)
        _, kwargs = mock_create_from_structure.call_args
        self.assertEqual(kwargs["notif_type"], NotificationType.TOWER_REFUELED_EXTRA)

    @patch(NOTIFICATIONS_PATH + ".Notification.send_to_webhook")
    def test_should_generate_refueled_notif_when_fuel_level_increased(
        self, mock_send_to_webhook
    ):
        # given
        structure = StructureFactory(
            owner=self.owner, fuel_expires_at=now() + dt.timedelta(hours=1)
        )
        old_instance = deepcopy(structure)
        # when
        structure.fuel_expires_at = now() + dt.timedelta(hours=12)
        structure.handle_fuel_notifications(old_instance)
        # then
        self.assertTrue(mock_send_to_webhook.called)

    @patch(NOTIFICATIONS_PATH + ".Notification.send_to_webhook")
    def test_should_not_generate_refueled_notif_when_fuel_level_almost_unchanged(
        self, mock_send_to_webhook
    ):
        # given
        target_date_1 = now() + dt.timedelta(hours=2)
        target_date_1 = now() + dt.timedelta(hours=2, minutes=15)  # FIXME
        structure = StructureFactory(owner=self.owner, fuel_expires_at=target_date_1)
        old_instance = deepcopy(structure)
        # when
        structure.fuel_expires_at = target_date_1
        structure.handle_fuel_notifications(old_instance)
        # then
        self.assertFalse(mock_send_to_webhook.called)

    @patch(NOTIFICATIONS_PATH + ".Notification.send_to_webhook")
    def test_should_not_generate_refueled_notif_fuel_level_decreased(
        self, mock_send_to_webhook
    ):
        # given
        structure = StructureFactory(
            owner=self.owner, fuel_expires_at=now() + dt.timedelta(hours=12)
        )
        old_instance = deepcopy(structure)
        # when
        structure.fuel_expires_at = now() + dt.timedelta(hours=1)
        structure.handle_fuel_notifications(old_instance)
        # then
        self.assertFalse(mock_send_to_webhook.called)

    @patch(NOTIFICATIONS_PATH + ".Notification.send_to_webhook")
    def test_should_not_generate_refueled_notif_fuel_is_removed(
        self, mock_send_to_webhook
    ):
        # given
        structure = StructureFactory(
            owner=self.owner, fuel_expires_at=now() + dt.timedelta(hours=2)
        )
        old_instance = deepcopy(structure)
        # when
        structure.fuel_expires_at = None
        structure.handle_fuel_notifications(old_instance)
        # then
        self.assertFalse(mock_send_to_webhook.called)


class TestStructurePowerMode(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        cls.owner = OwnerFactory()

    def test_returns_none_for_non_upwell_structures(self):
        starbase = StarbaseFactory.build(owner=self.owner)
        self.assertIsNone(starbase.power_mode)

        pos = PocoFactory.build(owner=self.owner)
        self.assertIsNone(pos.power_mode)

        structure = StructureFactory.build(owner=self.owner)
        self.assertIsNotNone(structure.power_mode)

    def test_full_power_mode(self):
        structure = StructureFactory.build(
            owner=self.owner, fuel_expires_at=now() + dt.timedelta(hours=1)
        )
        self.assertEqual(structure.power_mode, Structure.PowerMode.FULL_POWER)
        self.assertEqual(structure.get_power_mode_display(), "Full Power")

    def test_low_power_mode_1(self):
        structure = StructureFactory.build(
            owner=self.owner,
            fuel_expires_at=now() - dt.timedelta(seconds=3),
            last_online_at=now() - dt.timedelta(days=3),
        )
        self.assertEqual(structure.power_mode, Structure.PowerMode.LOW_POWER)

    def test_low_power_mode_2(self):
        structure = StructureFactory.build(
            owner=self.owner,
            fuel_expires_at=None,
            last_online_at=None,
            state=Structure.State.ANCHORING,
        )
        self.assertEqual(structure.power_mode, Structure.PowerMode.LOW_POWER)

    def test_low_power_mode_3(self):
        structure = StructureFactory.build(
            owner=self.owner,
            fuel_expires_at=None,
            last_online_at=now() - dt.timedelta(days=3),
        )
        self.assertEqual(structure.power_mode, Structure.PowerMode.LOW_POWER)
        self.assertEqual(structure.get_power_mode_display(), "Low Power")

    def test_abandoned_mode_1(self):
        structure = StructureFactory.build(
            owner=self.owner,
            fuel_expires_at=now() - dt.timedelta(seconds=3),
            last_online_at=now() - dt.timedelta(days=7, seconds=1),
        )
        self.assertEqual(structure.power_mode, Structure.PowerMode.ABANDONED)

    def test_abandoned_mode_2(self):
        structure = StructureFactory.build(
            owner=self.owner,
            fuel_expires_at=None,
            last_online_at=now() - dt.timedelta(days=7, seconds=1),
        )
        self.assertEqual(structure.power_mode, Structure.PowerMode.ABANDONED)
        self.assertEqual(structure.get_power_mode_display(), "Abandoned")

    def test_low_abandoned_mode_1(self):
        structure = StructureFactory.build(
            owner=self.owner,
            fuel_expires_at=now() - dt.timedelta(seconds=3),
            last_online_at=None,
        )
        self.assertEqual(structure.power_mode, Structure.PowerMode.LOW_ABANDONED)

    def test_low_abandoned_mode_2(self):
        structure = StructureFactory.build(
            owner=self.owner,
            fuel_expires_at=None,
            last_online_at=None,
        )
        self.assertEqual(structure.power_mode, Structure.PowerMode.LOW_ABANDONED)
        self.assertEqual(structure.get_power_mode_display(), "Abandoned?")


class TestStructureTags(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        cls.owner = OwnerFactory()
        EveSovereigntyMapFactory(
            eve_solar_system_name="1-PGSG", corporation=cls.owner.corporation
        )

    def test_can_create_generated_tags(self):
        # given
        obj = StructureFactory(owner=self.owner, eve_solar_system_name="1-PGSG")
        obj.tags.clear()
        # when
        obj.update_generated_tags()
        # then
        null_tag = StructureTag.objects.get(name=StructureTag.NAME_NULLSEC_TAG)
        self.assertIn(null_tag, list(obj.tags.all()))
        sov_tag = StructureTag.objects.get(name=StructureTag.NAME_SOV_TAG)
        self.assertIn(sov_tag, list(obj.tags.all()))

    def test_can_update_generated_tags(self):
        # given
        obj = StructureFactory(owner=self.owner, eve_solar_system_name="1-PGSG")
        null_tag = StructureTag.objects.get(name=StructureTag.NAME_NULLSEC_TAG)
        self.assertIn(null_tag, list(obj.tags.all()))
        null_tag.order = 100
        null_tag.style = StructureTag.Style.DARK_BLUE
        null_tag.save()

        sov_tag = StructureTag.objects.get(name=StructureTag.NAME_SOV_TAG)
        self.assertIn(sov_tag, list(obj.tags.all()))
        sov_tag.order = 100
        sov_tag.style = StructureTag.Style.RED
        sov_tag.save()

        # when
        obj.update_generated_tags(recreate_tags=True)

        # then
        null_tag.refresh_from_db()
        self.assertEqual(null_tag.style, StructureTag.Style.RED)
        self.assertEqual(null_tag.order, 50)
        sov_tag.refresh_from_db()
        self.assertEqual(sov_tag.style, StructureTag.Style.DARK_BLUE)
        self.assertEqual(sov_tag.order, 20)

    def test_can_handle_unknown_space_type_for_existing_tags(self):
        # given
        obj = StructureFactory(owner=self.owner, eve_solar_system_name="1-PGSG")
        obj.tags.clear()
        # when
        with patch(
            "structures.models.eveuniverse.EveSpaceType.from_solar_system"
        ) as mock:
            mock.return_value = EveSpaceType.UNKNOWN
            obj.update_generated_tags()

        # then
        self.assertEqual(obj.tags.count(), 1)
        sov_tag = StructureTag.objects.get(name=StructureTag.NAME_SOV_TAG)
        self.assertIn(sov_tag, list(obj.tags.all()))

    def test_can_handle_unknown_space_type_when_recreating_tags(self):
        # given
        obj = StructureFactory(owner=self.owner, eve_solar_system_name="1-PGSG")
        obj.tags.clear()
        # when
        with patch(
            "structures.models.eveuniverse.EveSpaceType.from_solar_system"
        ) as mock:
            mock.return_value = EveSpaceType.UNKNOWN
            obj.update_generated_tags(recreate_tags=True)

        # then
        self.assertEqual(obj.tags.count(), 1)
        sov_tag = StructureTag.objects.get(name=StructureTag.NAME_SOV_TAG)
        self.assertIn(sov_tag, list(obj.tags.all()))


class TestStructureTag(NoSocketsTestCase):
    def test_str(self):
        obj = StructureTag(name="Super cool tag")
        self.assertEqual(str(obj), "Super cool tag")

    def test_repr(self):
        obj = StructureTag.objects.create(name="Super cool tag")
        expected = "StructureTag(name='Super cool tag')"
        self.assertEqual(repr(obj), expected)

    def test_list_sorted(self):
        x1 = StructureTag(name="Alpha")
        x2 = StructureTag(name="charlie")
        x3 = StructureTag(name="bravo")
        tags = [x1, x2, x3]

        self.assertListEqual(StructureTag.sorted(tags), [x1, x3, x2])
        self.assertListEqual(StructureTag.sorted(tags, reverse=True), [x2, x3, x1])

    def test_html_default(self):
        x = StructureTag(name="Super cool tag")
        self.assertEqual(
            x.html, '<span class="label label-default">Super cool tag</span>'
        )

    def test_html_primary(self):
        x = StructureTag(name="Super cool tag", style="primary")
        self.assertEqual(
            x.html, '<span class="label label-primary">Super cool tag</span>'
        )


class TestStructureSave(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        cls.owner = OwnerFactory()
        EveSovereigntyMapFactory(
            eve_solar_system_name="1-PGSG", corporation=cls.owner.corporation
        )

    def test_can_save_tags_low_sec(self):
        obj = StructureFactory(owner=self.owner, eve_solar_system_name="Amamake")
        lowsec_tag = StructureTag.objects.get(name=StructureTag.NAME_LOWSEC_TAG)
        self.assertIn(lowsec_tag, obj.tags.all())
        self.assertIsNone(
            StructureTag.objects.filter(name=StructureTag.NAME_SOV_TAG).first()
        )

    def test_can_save_tags_null_sec_w_sov(self):
        obj = StructureFactory(owner=self.owner, eve_solar_system_name="1-PGSG")
        nullsec_tag = StructureTag.objects.get(name=StructureTag.NAME_NULLSEC_TAG)
        self.assertIn(nullsec_tag, obj.tags.all())
        sov_tag = StructureTag.objects.get(name=StructureTag.NAME_SOV_TAG)
        self.assertIn(sov_tag, obj.tags.all())

    def test_should_create_default_tags(self):
        # given
        default_tag = StructureTagFactory(is_default=True)
        # when
        obj = StructureFactory(owner=self.owner)
        # then
        self.assertIn(default_tag, obj.tags.all())

    def test_should_not_create_default_tags(self):
        # given
        default_tag = StructureTagFactory(is_default=True)
        # when
        obj = StructureFactory(owner=self.owner)
        obj.tags.all().delete()
        obj.save()
        # then
        self.assertNotIn(default_tag, obj.tags.all())


class TestStructureNoSetup(NoSocketsTestCase):
    def test_structure_get_matching_state(self):
        self.assertEqual(
            Structure.State.from_esi_name("anchoring"),
            Structure.State.ANCHORING,
        )
        self.assertEqual(
            Structure.State.from_esi_name("not matching name"),
            Structure.State.UNKNOWN,
        )

    def test_structure_service_get_matching_state(self):
        self.assertEqual(
            StructureService.State.from_esi_name("online"),
            StructureService.State.ONLINE,
        )
        self.assertEqual(
            StructureService.State.from_esi_name("offline"),
            StructureService.State.OFFLINE,
        )
        self.assertEqual(
            StructureService.State.from_esi_name("not matching"),
            StructureService.State.OFFLINE,
        )


class TestStructureService(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

    def test_str(self):
        structure = StructureFactory(
            eve_solar_system_name="Amamake", name="Test Structure Alpha"
        )
        obj = StructureServiceFactory(structure=structure, name="Clone Bay")
        expected = "Amamake - Test Structure Alpha - Clone Bay"
        self.assertEqual(str(obj), expected)



"""Managers for Structures."""

# pylint: disable=missing-class-docstring

from __future__ import annotations

import datetime as dt
import itertools
from typing import TYPE_CHECKING, Any, Iterable, Optional, Set, Tuple

from django.contrib.auth.models import User
from django.db import models, transaction
from django.db.models import Case, Count, Exists, OuterRef, Q, Sum, Value, When
from django.utils.timezone import now
from esi.models import Token
from eveuniverse.models import EveMoon, EvePlanet, EveSolarSystem, EveType

from allianceauth.eveonline.models import EveCorporationInfo
from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from . import __title__
from .app_settings import STRUCTURES_HOURS_UNTIL_STALE_NOTIFICATION
from .constants import EveCategoryId, EveTypeId
from .core.notification_types import NotificationType
from .providers import esi
from .webhooks.managers import WebhookBaseManager

if TYPE_CHECKING:
    from .models import Owner

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class EveSovereigntyMapManager(models.Manager):
    def update_or_create_all_from_esi(self):
        """Update or create complete sovereignty map from ESI."""
        sov_map = esi.client.Sovereignty.get_sovereignty_map().results()
        logger.info("Retrieved sovereignty map from ESI")
        last_updated = now()
        obj_list = []
        for solar_system in sov_map:
            obj_def = {
                "solar_system_id": solar_system["system_id"],
                "last_updated": last_updated,
            }
            for key in ["alliance_id", "corporation_id", "faction_id"]:
                if key in solar_system and solar_system[key]:
                    obj_def[key] = solar_system[key]
                else:
                    obj_def[key] = None

            if (
                obj_def["alliance_id"]
                or obj_def["corporation_id"]
                or obj_def["faction_id"]
            ):
                obj_list.append(self.model(**obj_def))

        if obj_list:
            logger.info("Storing sovereignty map ...")
            with transaction.atomic():
                self.all().delete()
                self.bulk_create(obj_list, batch_size=1000)

    def corporation_has_sov(
        self, eve_solar_system, corporation: EveCorporationInfo
    ) -> bool:
        """returns true if given corporation has sov in this solar system
        else False
        """
        if not eve_solar_system.is_null_sec:
            return False

        alliance_id = (
            int(corporation.alliance.alliance_id) if corporation.alliance else None
        )
        return bool(alliance_id) and (
            self.solar_system_sov_alliance_id(eve_solar_system) == alliance_id
        )

    def solar_system_sov_alliance_id(
        self, eve_solar_system: EveSolarSystem
    ) -> Optional[int]:
        """returns ID of sov owning alliance for this system or None"""
        if not eve_solar_system.is_null_sec:
            return None

        try:
            sov_map = self.get(solar_system_id=eve_solar_system.id)
        except self.model.DoesNotExist:
            return None

        return sov_map.alliance_id if sov_map.alliance_id else None


class NotificationBaseQuerySet(models.QuerySet):
    def annotate_can_be_rendered(self) -> models.QuerySet:
        """Annotate field indicating if a notification can be rendered."""

        return self.annotate(
            can_be_rendered_2=Case(
                When(notif_type__in=NotificationType.values, then=True),
                default=Value(False),
                output_field=models.BooleanField(),
            )
        )


class NotificationBaseManagerBase(models.Manager):
    def add_or_remove_timers(self):
        """Add or remove timers from notifications."""

        cutoff_dt_for_stale = now() - dt.timedelta(
            hours=STRUCTURES_HOURS_UNTIL_STALE_NOTIFICATION
        )
        notifications = (
            self.filter(notif_type__in=NotificationType.relevant_for_timerboard())
            .exclude(is_timer_added=True)
            .filter(timestamp__gte=cutoff_dt_for_stale)
            .select_related("owner")
            .order_by("timestamp")
        )
        if notifications.exists():
            for notification in notifications:
                notification.add_or_remove_timer()


class NotificationQuerySet(NotificationBaseQuerySet):
    pass


class NotificationManagerBase(NotificationBaseManagerBase):
    pass


NotificationManager = NotificationManagerBase.from_queryset(NotificationQuerySet)


class GeneratedNotificationQuerySet(NotificationBaseQuerySet):
    pass


class GeneratedNotificationManagerBase(NotificationBaseManagerBase):
    def get_or_create_from_structure(
        self, structure: models.Model, notif_type: NotificationType
    ) -> Tuple[Any, bool]:
        """Get or create an object from given structure."""

        if notif_type not in {NotificationType.TOWER_REINFORCED_EXTRA}:
            raise ValueError(f"Unsupported notification type: {notif_type}")

        return self._get_or_create_tower_reinforced(structure)

    def _get_or_create_tower_reinforced(self, structure) -> Tuple[Any, bool]:
        if not structure.is_starbase:
            raise ValueError(f"Structure is not a starbase: {structure}")
        if not structure.is_reinforced:
            raise ValueError(f"Starbase is not reinforced: {structure}")
        if not structure.state_timer_end:
            raise ValueError(f"Starbase has no reinforce time: {structure}")
        reinforced_until = structure.state_timer_end.isoformat()
        with transaction.atomic():
            try:
                obj = self.get(
                    structures=structure,
                    notif_type=NotificationType.TOWER_REINFORCED_EXTRA,
                    details__reinforced_until=reinforced_until,
                )
                created = False
            except self.model.DoesNotExist:
                obj = self.create(
                    owner=structure.owner,
                    notif_type=NotificationType.TOWER_REINFORCED_EXTRA,
                    details={"reinforced_until": reinforced_until},
                )
                obj.structures.add(structure)
                created = True
        return obj, created


GeneratedNotificationManager = GeneratedNotificationManagerBase.from_queryset(
    GeneratedNotificationQuerySet
)


class OwnerQuerySet(models.QuerySet):
    def annotate_characters_count(self) -> models.QuerySet:
        """Add character count annotations."""
        qs = self.annotate(
            characters_enabled_count=Count(
                "characters",
                filter=Q(
                    characters__character_ownership__isnull=False,
                    characters__is_enabled=True,
                ),
                distinct=True,
            )
        ).annotate(
            characters_disabled_count=Count(
                "characters",
                filter=Q(
                    characters__character_ownership__isnull=False,
                    characters__is_enabled=False,
                ),
                distinct=True,
            )
        )
        return qs

    def structures_last_updated(self) -> Optional[dt.datetime]:
        """Date/time when structures were last updated for any of the active owners."""
        obj = self.filter(is_active=True).order_by("-structures_last_update_at").first()
        if not obj:
            return None
        return obj.structures_last_update_at


class OwnerManagerBase(models.Manager):
    pass


OwnerManager = OwnerManagerBase.from_queryset(OwnerQuerySet)


class StructureQuerySet(models.QuerySet):
    def filter_upwell_structures(self) -> models.QuerySet:
        """Filter for upwell structures."""
        return self.filter(eve_type__eve_group__eve_category=EveCategoryId.STRUCTURE)

    def filter_customs_offices(self) -> models.QuerySet:
        """Filter for custom offices."""
        return self.filter(eve_type=EveTypeId.CUSTOMS_OFFICE)

    def filter_starbases(self) -> models.QuerySet:
        """Filter for starbases."""
        return self.filter(eve_type__eve_group__eve_category=EveCategoryId.STARBASE)

    def filter_skyhooks(self) -> models.QuerySet:
        """Filter for skyhooks."""
        return self.filter(eve_type=EveTypeId.ORBITAL_SKYHOOK)

    def ids(self) -> Set[int]:
        """Return ids as set."""
        return set(self.values_list("id", flat=True))

    def select_related_defaults(self) -> models.QuerySet:
        """returns a QuerySet with the default select_related"""
        return self.select_related(  # pylint: disable=E1102
            "owner",
            "owner__corporation",
            "owner__corporation__alliance",
            "eve_type",
            "eve_solar_system",
            "eve_solar_system__eve_constellation__eve_region",
            "eve_planet",
            "eve_moon",
            "eve_type__eve_group",
            "eve_type__eve_group__eve_category",
        )

    def visible_for_user(self, user: User) -> models.QuerySet:
        """Return structures which the given user have permission to view."""
        if user.has_perm("structures.view_all_structures"):
            return self

        if user.has_perm("structures.view_corporation_structures") or user.has_perm(
            "structures.view_alliance_structures"
        ):
            corporation_ids = set(
                user.character_ownerships.values_list(
                    "character__corporation_id", flat=True
                )
            )
            corporations = list(
                EveCorporationInfo.objects.select_related("alliance").filter(
                    corporation_id__in=corporation_ids
                )
            )
        else:
            corporations = []

        if user.has_perm("structures.view_alliance_structures"):
            alliances = {
                corporation.alliance
                for corporation in corporations
                if corporation.alliance
            }
            for alliance in alliances:
                corporations += list(alliance.evecorporationinfo_set.all())

            corporations = list(set(corporations))

        return self.filter(owner__corporation__in=corporations)

    def filter_tags(self, tag_names: Iterable[str]) -> models.QuerySet:
        """Apply filter for given tags."""
        if not tag_names:
            return self

        query = self.filter(tags__name__in=list(tag_names)).distinct()
        return query

    def annotate_has_poco_details(self) -> models.QuerySet:
        """Add annotation wether the structure has poco details."""
        from .models import PocoDetails

        return self.annotate(
            has_poco_details=Exists(
                PocoDetails.objects.filter(structure_id=OuterRef("id"))
            )
        )

    def annotate_has_starbase_detail(self) -> models.QuerySet:
        """Add annotation whether the structure has starbase details."""
        from .models import StarbaseDetail

        return self.annotate(
            has_starbase_detail=Exists(
                StarbaseDetail.objects.filter(structure_id=OuterRef("id"))
            )
        )

    def annotate_jump_fuel_quantity(self) -> models.QuerySet:
        """Add annotation with quantity of jump fuel."""
        from .models import StructureItem

        return self.annotate(
            jump_fuel_quantity_2=Sum(
                "items__quantity",
                filter=Q(
                    items__eve_type=EveTypeId.LIQUID_OZONE,
                    items__location_flag=StructureItem.LocationFlag.STRUCTURE_FUEL,
                ),
            )
        )


class StructureManagerBase(models.Manager):
    def get_or_create_esi(self, *, id: int, token: Token) -> Tuple[Any, bool]:
        """get or create a structure with data from ESI if needed.

        Args:
            id: Structure ID of object in Eve Online
            token: ``esi.models.Token`` object with scope:
                ``esi-universe.read_structures.v1``

        Returns:
            object, created
        """
        id = int(id)
        try:
            obj = self.get(id=id)
            return obj, False
        except self.model.DoesNotExist:
            return self.update_or_create_esi(id=id, token=token)

    def update_or_create_esi(self, *, id: int, token: Token) -> Tuple[Any, bool]:
        """update or create a structure from ESI for given structure ID
        This will only fetch basic info about a structure

        Args:
            id: Structure ID of object in Eve Online
            token: ``esi.models.Token`` object with scope: ``esi-universe.read_structures.v1``

        Returns:
            object, created
        """
        from .models import Owner

        id = int(id)
        logger.info("Trying to fetch structure from ESI with ID %s", id)
        if token is None:
            raise ValueError("Can not fetch structure without token")

        structure_info = esi.client.Universe.get_universe_structures_structure_id(
            structure_id=id, token=token.valid_access_token()
        ).results()
        structure = {
            "structure_id": id,
            "name": self.model.extract_name_from_esi_response(structure_info["name"]),
            "position": structure_info["position"],
            "type_id": structure_info["type_id"],
            "system_id": structure_info["solar_system_id"],
        }
        owner = Owner.objects.get(
            corporation__corporation_id=structure_info["corporation_id"]
        )
        obj, created = self.update_or_create_from_dict(structure=structure, owner=owner)
        return obj, created

    def update_or_create_from_dict(
        self, structure: dict, owner: Owner
    ) -> Tuple[Any, bool]:
        """update or create structure from given dict"""

        eve_type: EveType = EveType.objects.get_or_create_esi(id=structure["type_id"])[
            0
        ]
        eve_solar_system, _ = EveSolarSystem.objects.get_or_create_esi(
            id=structure["system_id"]
        )
        position_x, position_y, position_z = self._extract_position(structure)
        eve_planet = self._extract_eve_planet(structure)
        eve_moon = self._extract_eve_moon(structure)

        try:
            old_obj = self.get(id=structure["structure_id"])
        except self.model.DoesNotExist:
            old_obj = None

        obj, created = self.update_or_create(
            id=structure["structure_id"],
            defaults={
                "owner": owner,
                "eve_type": eve_type,
                "name": structure.get("name", ""),
                "eve_solar_system": eve_solar_system,
                "eve_planet": eve_planet,
                "eve_moon": eve_moon,
                "position_x": position_x,
                "position_y": position_y,
                "position_z": position_z,
                "fuel_expires_at": structure.get("fuel_expires"),
                "next_reinforce_hour": structure.get("next_reinforce_hour"),
                "next_reinforce_apply": structure.get("next_reinforce_apply"),
                "reinforce_hour": structure.get("reinforce_hour"),
                "state": self.model.State.from_esi_name(structure.get("state", "")),
                "state_timer_start": structure.get("state_timer_start"),
                "state_timer_end": structure.get("state_timer_end"),
                "unanchors_at": structure.get("unanchors_at"),
                "last_updated_at": now(),
            },
        )

        if old_obj:
            obj.handle_fuel_notifications(old_obj)

        # Make sure we have dogmas loaded with all upwell structures for fittings
        if eve_type.eve_group.eve_category_id == EveCategoryId.STRUCTURE:
            EveType.objects.get_or_create_esi(
                id=structure["type_id"], enabled_sections=[EveType.Section.DOGMAS]
            )

        self._save_related_structure_services(structure, obj)

        return obj, created

    def _extract_eve_moon(self, structure):
        if moon_id := structure.get("moon_id"):
            return EveMoon.objects.get_or_create_esi(id=moon_id)[0]
        return None

    def _extract_eve_planet(self, structure):
        if planet_id := structure.get("planet_id"):
            return EvePlanet.objects.get_or_create_esi(id=planet_id)[0]
        return None

    def _extract_position(self, structure):
        if position := structure.get("position"):
            return position.get("x"), position.get("y"), position.get("z")
        return None, None, None

    @staticmethod
    def _save_related_structure_services(structure, obj):
        from .models import StructureService

        StructureService.objects.filter(structure=obj).delete()
        if "services" in structure and structure["services"]:
            for service in structure["services"]:
                service_state = StructureService.State.from_esi_name(service["state"])
                args = {
                    "structure": obj,
                    "name": service["name"],
                    "state": service_state,
                }
                StructureService.objects.create(**args)

        if obj.services.filter(state=StructureService.State.ONLINE).exists():
            obj.last_online_at = now()
            obj.save()


StructureManager = StructureManagerBase.from_queryset(StructureQuerySet)


class StructureTagManager(models.Manager):
    def get_or_create_for_space_type(
        self, solar_system: EveSolarSystem
    ) -> Tuple[Any, bool]:
        """Get or create tag for a space type."""
        params = self._obj_params_from_solar_system(solar_system)
        try:
            obj = self.get(name=params["name"])
            return obj, False
        except self.model.DoesNotExist:
            return self.update_or_create_for_space_type(solar_system)

    def update_or_create_for_space_type(
        self, solar_system: EveSolarSystem
    ) -> Tuple[Any, bool]:
        """Update or create tag for a space type."""
        params = self._obj_params_from_solar_system(solar_system)
        return self.update_or_create(
            name=params["name"],
            defaults={
                "style": params["style"],
                "description": ("this tag represents a space type. system generated."),
                "order": 50,
                "is_user_managed": False,
                "is_default": False,
            },
        )

    def _obj_params_from_solar_system(self, solar_system):
        from .models import EveSpaceType

        space_type = EveSpaceType.from_solar_system(solar_system)
        params = self.model.SPACE_TYPE_MAP.get(space_type)
        if not params:
            raise ValueError("Unknown space type")
        return params

    def get_or_create_for_sov(self) -> Tuple[Any, bool]:
        """Get or create sovereignty tag."""
        try:
            obj = self.get(name=self.model.NAME_SOV_TAG)
            return obj, False
        except self.model.DoesNotExist:
            return self.update_or_create_for_sov()

    def update_or_create_for_sov(self) -> Tuple[Any, bool]:
        """Update or create sovereignty tag."""
        return self.update_or_create(
            name=self.model.NAME_SOV_TAG,
            defaults={
                "style": self.model.Style.DARK_BLUE,
                "description": (
                    "Owner of this structure has sovereignty. system generated."
                ),
                "order": 20,
                "is_user_managed": False,
                "is_default": False,
            },
        )


class WebhookManager(WebhookBaseManager):
    def enabled_notification_types(self) -> Set[str]:
        """Set of all currently enabled notification types."""
        notif_types_list = list(
            self.filter(is_active=True).values_list("notification_types", flat=True)
        )
        return set(itertools.chain(*notif_types_list))
